import importlib
import json
import logging
import os
import os.path
import sys
import time
from datetime import datetime
from pathlib import Path

log = logging.getLogger(__name__)

class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'

if sys.version_info.minor < 12:
    import imp
    def importModule(moduleName, moduleFile):
        """
        Imports a module from source code and returns the exected module
        @param moduleName Name of the module
        @param moduleFile File of the source code
        @return Excecuted module is returned 
        @throws Exception in case of an error
        """
        log.debug("=> importModule(moduleName=%s, moduleFile=%s)" % (moduleName, moduleFile))
        return imp.load_source(moduleName, moduleFile)
else:
    def importModule(moduleName, moduleFile):
        """
        Imports a module from source code and returns the exected module
        @param moduleName Name of the module
        @param moduleFile File of the source code
        @return Excecuted module is returned 
        @throws Exception in case of an error
        """
        log.debug("=> importModule(moduleName=%s, moduleFile=%s)" % (moduleName, moduleFile))
        return importlib.machinery.SourceFileLoader(moduleName, moduleFile).load_module()

def newInstanceFromClass(moduleName, className):
    '''
    Creates an instance of a class from a module and class name string.
    @param moduleName Name of the module
    @param className Name of the class in the module.
    @return Instance of class
    @throws Exception in case of an error.
    '''
    log.debug("=> createClassInstance(moduleName=%s, className=%s)" % (moduleName, className))

    mod = importlib.import_module(moduleName)
    cl = getattr(mod, className, None)

    if cl == None:
        raise Exception("Could not get class object %s from module %s." % (className, moduleName))

    obj = cl()

    if obj == None :
        raise Exception("Could not create instance of class %s from module %s." % (className, moduleName))
    
    log.debug("<= createClassInstance()");
    return obj

def splitString(str, sepChar):
    a = []
    for s in str.split(sepChar):
        a.append(s.strip())
    return a

def isDict(obj):
    if not obj:
        return False
    else:
        return True if obj.__class__ == {}.__class__ else False

def isArray(obj):
    if not obj:
        return False
    else:
        return True if obj.__class__ == [].__class__ else False

def isString(obj):
    return obj.__class__ == "str".__class__ if obj else False

def convert_str_to_python_type(s:str):
    if not s:
        return s

    try:
        return int(s)
    except Exception:
        pass
    
    try:
        return float(s)
    except Exception:
        pass
    
    if s.lower() == 'true':
        return True
    elif s.lower() == 'false':
        return False

    try:
        result = json.loads(s)
        if isinstance(result, (list, dict, tuple)):
            return result
    except json.JSONDecodeError:
        pass            
    
    return s

#MARK: execCmd
from enum import Enum


class ReturnType(Enum):
    TEXT = 1
    JSON_AS_DICT = 2
    LINES_AS_LIST = 3

def execCmd(cmd, returnType : ReturnType = ReturnType.TEXT):
    log.debug("Exec %s" % (cmd))
        
    cmdIn = os.popen(cmd)

    line = cmdIn.readline()
    s = ""
    while line:
        if line.startswith("b\'"):
            line = line.strip()
            line = line[2:-1] + "\n"
        s = s + line
        line = cmdIn.readline()    

    log.debug("\n%s" % (s.strip()))                                  

    if returnType == ReturnType.JSON_AS_DICT:
        return json.loads(s)
    elif returnType == ReturnType.LINES_AS_LIST:
        return splitString(s, "\n")
    else:
        return s

#MARK: readlines
def readlines(filepath:str="-", 
              stripLines:bool=False, 
              skipEmptyLines:bool=False, 
              skipLinesStartWith:list[str]=[],
              encoding="utf-8"):
    '''
    Generator function to read a file linewise.
    You can use this function to read the file line by line with the following code:
        
    lines = readlines(filename)
    while (line := next(lines)):
        print(line) 
        
    Advanage of this function is that file is not read into the memory completely. By this unlimited file sizes
    can be processed.
    '''
    if not filepath or filepath == "-":
        file = sys.stdin
        iamOwner = False
    else:        
        file = open(filepath, "r", encoding=encoding)
        iamOwner = True

    while True:
        line = file.readline()
        
        # If we reached end of file
        if line == '':
            if iamOwner:
                file.close()
            yield None

        if stripLines:
            line=line.strip()
        else:
            line=line.rstrip()
            
        if skipEmptyLines and len(line)==0:
            continue

        if len(line)>0 and line[0] in skipLinesStartWith:
            continue

        yield line    
class FileLinesIterator:
    '''
    Iterator class to walk through lines of a text file in the context of for loop.

    Sample code:    

    from file_lines_iter import FileLines

    for line in FileLinesIterator("/Users/rolan/Desktop/WORK/1646599091.json"):
        print(line)
        
    '''
    
    def __init__(self, filepath, stripLines:bool=False, skipEmptyLines:bool=False, skipLinesStartWith:list[str]=[]):
        '''
        Constructor of FileLinesIterator class.
        :param filepath: Path of the file through which lines shall be iterated. If "-" than stdin is read
        :param skipLines: If set all lines are left and right stripped.
        :param skipEmptyLines: If set to true empty lines are not returned.
        :param skipLinesStartWith: Array of single characters. If a line starts with one of the chars it is not returned.
        '''
        self._filepath = filepath
        self._file = None
        self._skipEmptyLines=skipEmptyLines
        self._skipLinesStartWith=skipLinesStartWith
        self._stripLines=stripLines

    def __iter__(self):
        return self
    
    def __next__(self):
        if not self._file:
            if self._filepath=="-":
                self._file = sys.stdin
            else:
                self._file = open(self._filepath, "r")

        while True:
            line = self._file.readline()
            if not line:
                if self._filepath!="-":
                    self._file.close()
                raise StopIteration

            if self._stripLines:
                line=line.strip()
            else:
                line=line.rstrip()
                
            if self._skipEmptyLines and len(line)==0:
                continue

            if len(line)>0 and line[0] in self._skipLinesStartWith:
                continue

            return line

import socket


def getIp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

# Class to executues a function every x seconds 
import threading
import time
import traceback


class FunctionExecutor(threading.Thread):
    def __init__(self, cronSecs, function, *args, **kwargs):
        super().__init__()
        self._cronSecs = cronSecs
        self._function = function
        self._args = args
        self._kwargs = kwargs

    def run(self):
        while True:
            try:
                time.sleep(self._cronSecs)
                self._function(*self._args, **self._kwargs)
            except Exception as ex:
                log.error(str(ex))
                traceback.print_exc()
                
def getValueFromDict(rootD:dict, elemPath:str, default:any=None, sepChar=".") -> any:
    '''
    Gets a value from a dictionary element.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    name = names[0]    

    if name not in d.keys() or len(names) == 0:
        return default
    elif len(names) == 1:
        return d.get(name)            
    else:
        return getValueFromDict(d.get(names[0]), sepChar.join(names[1:]), default=default, sepChar=sepChar)

def get_value_from_dict_2(rootD:dict, elemPath:str, default:any=None, sepChar=".") -> any:
    '''
    New version of getValueFromDict which supports arrays.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    name = names[0]    

    # If name represents an elemement in an array such as dvi[0]
    list_idx = -1
    
    if (idx1 := name.find("[")) > -1:
        if (idx2 := name.find("]", idx1)) == -1:
            return default
        list_idx = int(name[idx1+1:idx2])
        name = name[0:idx1]
        
    if name not in d.keys() or len(names) == 0:
        return default
    elif len(names) == 1:
        if list_idx == -1:
            return d.get(name)            
        else:
            l = d.get(name)
            return l[list_idx] if list_idx < len(l) else default
    else:
        if list_idx == -1:
            return get_value_from_dict_2(d.get(names[0]), sepChar.join(names[1:]), default=default, sepChar=sepChar)
        else:
            l = d.get(name)
            if list_idx >= len(l):
                return default
            return get_value_from_dict_2(l[list_idx], sepChar.join(names[1:]), default=default, sepChar=sepChar)
            

def has_element_value(d:dict, elem_name_path:str, test_value)->bool:
    """Verifies if an element in a dictionary has a specific value. 
       By using the function get_value_from_dict_2 it can support arrays in the elem_name_path.
    Returns:
        bool: True if the element with the value exists. 
    """
    _DOES_NOT_EXIST = "__DOES_NOT_EXIST__"
    
    v = get_value_from_dict_2(d, elem_name_path, default=_DOES_NOT_EXIST, sepChar=".")
    
    if v != None and v == _DOES_NOT_EXIST:
        return False        
    elif v == None and test_value == None:
        return True
    elif test_value != None and v == None:
        return False
    elif test_value == None and v != None:
        return False
    # Check if the searched value is in the array
    elif type(v) in [list, tuple] and type(test_value) not in [list, tuple]:
        return test_value in v
    # If the found value and test value have not the same type, they are not equal.
    elif type(v) != type(test_value):
        return False
    else:        
        return v == test_value
    
def hasElementInDict(rootD:dict,     # Dictionary which shall be checked for the element
                     elemPath:str,   # Element path string
                     sepChar="."     # Dictionary level separator charactor in element path string.
) -> bool:                           # True if the element exists. False if the element doesn't exist.
    '''
    Verifies if a specific element in the dictionary exists by parsing an element path string.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    firstName = names[0]    

    if firstName not in d.keys() or len(names) == 0:
        return False
    elif firstName not in d.keys():
        return False
    elif len(names) == 1:
        return True if firstName in d.keys() else False          
    else:
        return hasElementInDict(d.get(names[0]), sepChar.join(names[1:]), sepChar=sepChar)

def setValueInDict(rootD:dict,              # Dictionary in which the value of an element shall be set.
                   elemPath:str,            # Path string to indicate the element in dictionary
                   value,                   # Value to set.
                   sepChar=".",             # Character which seperates the dictionary levels in path string
                   createIfNotExist=False   # If true, the element is created if it doesn't exist.
) -> bool:                                  # True if the value has been set. False if the value
                                            # hasn't been set because the element hasn't been found.
    '''
    Sets a value to an element in a dictionary by parsing an element path.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    firstName = names[0]    

    if firstName not in d.keys():
        if createIfNotExist:
            if len(names) > 1:
                for i in range(0, len(names) - 1):
                    d[names[i]] = {}
                    d = d[names[i]]
            d[names[-1]] = value
            return True
        else:
            return False

    # At this point the element with firstName always exist.
    # If we reached the leaf of the path
    if len(names) == 1:
        d[firstName] = value
        return True
    else:
        return setValueInDict(d.get(names[0]), sepChar.join(names[1:]), value, sepChar=sepChar, createIfNotExist=createIfNotExist)

def flattenDict(inD:dict,       # Multilevel dictionary which elements shall be flatten to 
                                # to a single level dictionary.
                prefix:str="",  # Prefix which shall be added to each element
                sepChar:str="_" # Character which shall be used to seperate the level from each other
)->dict:                        # Single level dirctionary
    '''
    Walks top/down through a multi-level dictionary and creates out of it a dictionary
    with only a single level. The names of the sub-dictionary in the input dictionary
    are separated in the name of the return dictionary by a separator character.
    '''
    retD = {}
    for k in inD.keys():
        e = inD[k]
        if type(e) == dict:
            d = flattenDict(e, prefix=f"{prefix}{k}{sepChar}", sepChar=sepChar)
            retD.update(d)
        else:
            retD[f"{prefix}{k}"] = inD[k]

    return retD

def suppressEx(func, *args, **kwargs) -> any:
    '''
    Runs a function and suppresses an exception. 
    
    :param *args: Optional arguments to pass into the function.
    :param **kwargs: Option keyword arguments to pass into the function.
    
    :return: Return of the function or None in case an exception is caught.
    '''
    try:
        return func(*args, **kwargs)
    except Exception as ex:
        log.error(f"Error while exec {func.__name__}: {ex}")        
        return None           
    
def int2StrWithDots(
    v:int,      # integer
    default:str=""    # default value 
)->str:
    '''
    Converts an integer to a string with "." as decimal seperator.
    '''
    if not v:
        return default
    else:
        return f"{v:,}".replace(",", ".")    

###
# Some usefull decorators
#MARK: profile decorator
from functools import wraps

PROFILE_SWITCHED_ON=False
def profile(log_func=None,       # Logger function to use for output. 
                                 # If not defined "print" is used to output to stdout.
            log_prefix="profile" # Prefix string for the log output
):                               # Any value of decorated function
    '''
    Function decorator to print function call arguments and execution time to stdout or a logger.
    
    By default the profile decoration is switchd off. To switch it on set the variable 
    utils.PROFILE_SWTICHED_ON=True.
    
    How to use the profile decorator:

    import utils
    
    utils.PROFILE_SWITCHED_ON=true
    
    @utils.profile(logger=log.info)
    def myFunc(name):
        print(f"Hallo from {name}")
    '''
    
    if not log_func:
        log_func = print
        
    def decorator(fu):
        if not PROFILE_SWITCHED_ON:
            return fu
        @wraps(fu)
        def wrapper(*args, **kwargs):
            s = ""
            for a in args:
                if s:
                    s += ", "
                s += f"{a}" 
            for (k, v) in kwargs.items():
                if s:
                    s += ", "
                s += f"{k}={v}" 
            log_func(f"{log_prefix}: calling {fu.__name__}({s})")
                
            before = time.time()
            rv = fu(*args, **kwargs)
            after = time.time()
            
            log_func(f"{log_prefix}: called {fu.__name__}: {after - before:.6f} secs")

            return rv
        return wrapper
    return decorator        


UTF8_to_UNICODE_ESCAPE = lambda s: s.encode('unicode_escape').decode('ascii')
UNICODE_ESCAPE_to_UTF8 = lambda s: s.encode('ascii').decode('unicode_escape')

def xxcode_str_values_in_dict(d:dict,   # Dictionary in which the string values shall be en- or de-coded. 
                              xxcoder   # Lambda function to en- or de-code.
                             ) -> None:
    '''
    En- or de-codes the string values within a dictionary recursivly.
    '''
    if not d:
        return None

    for k, v in d.items():
        if type(v) == str:
            d[k] = xxcoder(v)
        elif type(v) == dict:
            xxcode_str_values_in_dict(v, xxcoder)

    return None

########################################################################
#MARK: date function
########################################################################
from datetime import datetime


def fnow(format:str="%Y-%m-%d %H:%M:%S")->str:
    return datetime.now().strftime(format)

class DatetimeJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        else:
            return super().default(obj)   

########################################################################
#MARK: base64
########################################################################
import base64


def str_to_b64str(s:str)->str:
    if not s:
        return s
    b = s.encode('utf-8')
    b64 = base64.b64encode(b)
    return b64.decode('utf-8')    

def to_basic_auth(username:str, password:str)->str:
    return str_to_b64str(f"{username}:{password}")

########################################################################
#MARK: retry_function
########################################################################
def retry_function(retry_function, args:list=None, kwargs:dict=None, 
                   max_try_count:int=-1, max_secs:int=-1,
                   sleep_before_first_try_secs:int=-1, sleep_after_error_secs:int=-1,
                   error_return_value=None, 
                   info_message=None,
                   info_logger=None, 
                   error_logger=None):
    
    il = info_logger or log.info
    el = error_logger or log.error
        
    start_secs = time.time()

    last_ex = None    
    try_count = 0
    
    if info_message:
        il(info_message)
        
    if sleep_before_first_try_secs > 0:
        time.sleep(sleep_before_first_try_secs)
        
    while True:
        try_count += 1
        
        # Test if the max duration of the max number of trys have been reached
        now_secs = time.time()
        if (now_secs - start_secs > max_secs) or try_count > max_try_count:
            if last_ex:
                raise last_ex
            else:
                return error_return_value

        # Execute the function
        try:
            if try_count > 1 and info_message:
                il(f"... {try_count-1}. retry ...")
            
            rv = retry_function(*args, **kwargs)

            return rv
        
        except Exception as ex:
            el(ex)
            last_ex = ex
            if sleep_after_error_secs > 0 and try_count < max_try_count:
                log.info(f"... sleeping {sleep_after_error_secs} seconds before retrying ...")
                time.sleep(sleep_after_error_secs)

    
########################################################################
#MARK: enumerations
########################################################################
    
def get_enum_name_by_value(enum_class, value, default=None)->str | None:
    for member in enum_class:
        if member.value == value:
            return member.name
    return default

def get_enum_value_by_name(enum_class, name:str, default=None)->str | None:
    for member in enum_class:
        if member.name == name:
            return member.value
    return default


def get_enum_by_name(enum_class, name)->Enum | None:
    for member in enum_class:
        if member.name == name:
            return member
    return None

########################################################################
#MARK: UniqueList
########################################################################
class UniqueList:
    """List class to deal with unique elements.
    """
    def __init__(self, iterable=None):
        self._list = list()
        self._dict = dict() # Dictionary to manage items as keys to speadup 
                            # check if the item already exists in list
        
        if iterable:
            for item in iterable:
                self.append(item)

    @property 
    def list(self)->list:
        return self._list
        
    def append(self, item):
        if not item:
            raise ValueError(f"Empty items not allowed in list with unique values.")()
        if item not in self._dict.keys():
            self._list.append(item)
            self._dict[item] = True

    def extend(self, iterable):
        for item in iterable:
            self.append(item)

    def remove(self, item):
        if item in self._dict.keys():
            self._list.remove(item)
            del self._dict[item]

    def __contains__(self, item):
        return item in self._dict.keys()

    def __len__(self):
        return len(self._list)

    def __getitem__(self, index):
        return self._list[index]

    def __setitem__(self, index, value):
        if value in self._list and self._list[index] != value:
            raise ValueError(f"{value} already exists in UniqueList")
        self._list[index] = value
        del self._dict[value]
        self._dict[value] = True
        
    def __delitem__(self, index):
        del self._dict[self._list[index]]
        del self._list[index]

    def __iter__(self):
        return iter(self._list)

    def __repr__(self):
        return f"UniqueList({self._list})"

    def clear(self):
        self._list = list()
        self._dict = dict()

def integrate_lists(list_1:list, list_2:list)->list:
    """Integrates a list-2 with a list-1 or replaces a list-1 by a list-2. 

    Args:
        list_1 (list): List-1 which elements shall be updated, deleted or replaced completely by the the elements of list-2
        list_2 (list): List-2 which elements shall integrated with list-1.

    Returns:
        list: New list
    """

    # Determine if we need to perform a replace or an integration operation
    operation = "r"  # r := replace
    del_elems = {}
    add_elems = {}
        
    for e in list_2:
        if e[0] == "+":
            operation = "i"
            add_elems[e[1:]] = True
        elif e[0] == "-":
            operation = "i"
            del_elems[e[1:]] = True

    if operation == "r":
        return list_2
    
    ul = UniqueList(list_1)
    
    for e in list_2:
        #print(f"e: {e}")
        if e[0] == "-":
            # If the complete list shall be completely removed
            if e == "-*":
                return ()
            else:
                ul.remove(e[1:])
        elif e[0] == "+":
            ul.append(e[1:])
        else:
            ul.append(e)

    return ul.list        

########################################################################
#MARK: UniqueList
########################################################################

import argparse


def argparse_type_list(arg_value):
    try:
        if arg_value[0] == "\\":
            arg_value = arg_value[1:]

        if arg_value[0] != "[":
            return splitString(arg_value, ",")
        else:
            o = json.loads(arg_value)
    
        if type(o) != list:
            raise argparse.ArgumentTypeError(f"Not a list format: {arg_value}")
        
        return o
    
    except json.JSONDecodeError:
        raise argparse.ArgumentTypeError(f"Not a valid JSON list format: {arg_value}")

def argparse_type_json(arg_value):
    try:
        return json.loads(arg_value)
    except json.JSONDecodeError:
        raise argparse.ArgumentTypeError(f"Not a valid JSON format: {arg_value}")








#!/usr/bin/env python3

import json
import logging
import traceback
from typing import List

import uvicorn
from fastapi import (APIRouter, Body, Depends, FastAPI, Header, HTTPException,
                     Path, Query, Request, Response, Security, status)
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi.security.api_key import APIKeyHeader

import dtiot_d2c.cli.prfmanapi.api_impl.healthz_get as healthz_get
import dtiot_d2c.cli.prfmanapi.api_impl.profile_get as profile_get
import dtiot_d2c.cli.prfmanapi.config as CFG
from dtiot_d2c.cli.prfmanapi.api_impl.datamodel import HealthzResponse

log = logging.getLogger(__name__)


if __name__ == '__main__':

    logging.basicConfig(force=True, level=CFG.LogLevel.upper(),
                        format="%(asctime)s | %(levelname)s | %(threadName)s | %(name)s | %(message)s")                                

    try:
        # Create the API server     

        api_app = FastAPI(tite=CFG.ServiceName, 
                          include_in_schema=True, 
                          redirect_slashes=False,
                          openapi_url="/openapi.json",
                          docs_url=CFG.DocsUrl,
                          redoc_url=None,
                          title="Device to Cloud CLI profile management API",
                          description="API manage Device to Cloud CLI profiles under ~/.d2c/ directory.",
                          version=CFG.AppVersion,
                          servers=[
                            {"url": CFG.ExternalUrl, "description": f"{CFG.NAME}"},
                          ],
                          openapi_tags=[
                              {
                                  "name":"General",
                                  "description":"General endpoints"
                              },
                              {
                                  "name":"Profiles",
                                  "description":"Profile related endpoints"
                              }
                          ])
        log.info(f"You can access swagger doc at {CFG.ExternalUrl}{CFG.DocsUrl}")

        api_app.add_middleware(CORSMiddleware,
                               allow_origins=["*"],
                               allow_credentials=False,
                               allow_methods=["*"],
                               allow_headers=["*"])

        #########################################
        # General
        #########################################
        
        ### GET {{baseUrl}}/healthz
        @api_app.get(
            "/healthz",
            tags=["General"],
            summary="Health test of the endpoint.",
            description="Performs a health test of the endpoint."
        )
        async def _healthz_get(request:Request)->HealthzResponse:
            #print("dadsfasdfasdfasdfasdfasdfasdfasd")
            return healthz_get.handleRequest(request)

        #########################################
        # Profile
        #########################################

        ### GET {{baseUrl}}/profile
        @api_app.get(
            "/profile",
            tags=["Profiles"],
            summary="Gets all profiles",
            description="Gets all profiles.",
            response_description="List of profile JSONs."
        )
        async def _profile_get(request:Request)->List[dict]:
            return profile_get.handleRequest(request)

        ### GET {{baseUrl}}/profile/{{profileId}}         
        @api_app.get(
            "/profile/{id}",
            tags=["Profiles"],
            summary="Get details of a profile",
            description="Gets the detail json of a profile.",
            response_description="Profile JSON"
        )
        async def _profile_get_by_id(request:Request, id:str = Path(description="Id of the profile."),
                    )->dict:
            return profile_get.handleRequest(request, id=id)


        #########################################
        # Reports
        #########################################
        
        ### POST {{baseUrl}}/device-report/{{tenantId}}?when=last-month
        # when_desc = f"this-hour | last-hour | this-day | last-day | this-month | last-month | month-of-year | from-to"
        # month_of_year_desc = f"if when==month-of-year: MM.YYYY"
        # from_date_desc = f"if when==from-to: DD.MM.YYYY hh:mm:ss"
        # to_date_desc = f"if when==from-to: DD.MM.YYYY hh:mm:ss"
        
        # @api_app.post(
        #     "/device-report/{tenant_id}",
        #     tags=["Reports"],
        #     summary="Starts a device report",
        #     description="Creates a starts a ReportJob to create a device report for a specific reporting intervall.",
        #     response_description="Created job from which the processing status can be queried and the report file be downloaded"
        # )
        # async def _device_report_post(request:Request, 
        #                         tenant_id:int = Path(description="Id of the IMPACT tenant for which branch a report shall be created."),
        #                         when: str = Query(None, description=when_desc),
        #                         month_of_year:str = Query(None, description=month_of_year_desc),
        #                         from_date:str= Query(None, description=from_date_desc),
        #                         to_date:str= Query(None, description=to_date_desc),
        #                         token_info:TokenInfo=Depends(verify_token))->ReportJob:
        #     return device_report_post.handleRequest(token_info, request, tenant_id, when, 
        #                                             month_of_year, from_date,to_date)

        ####
        # Put the url path prefix in fron all endpoints
        
        main_app = FastAPI(openapi_url=None, #"/openapi.json",
                           docs_url=f'/docs',
                           redoc_url=None,
                          )
        main_app.mount(f"{CFG.ServiceUrlPrefix}", api_app)
        
        #########################################
        # RUN
        #########################################

        uvicorn.run(main_app,
                    host="0.0.0.0", 
                    port=int(CFG.ServicePort),
                    log_level="info",
                    access_log=False,
                    reload=False,
                    workers=1)

    except Exception as ex:
        log.error("ERROR: %s" % (str(ex)))
        traceback.print_exc()




import json
import logging
import sys
import time
from typing import List

import dtiot_d2c.d2c.test.utils as tutils
import dtiot_d2c.dmo as dmo
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.application import Application
from dtiot_d2c.d2c.device import Device
from dtiot_d2c.d2c.device.consts import *
from dtiot_d2c.d2c.device_group import DeviceGroup
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.d2c.test.application import (TestCase_CreateApplication,
                                            delete_application)
from dtiot_d2c.d2c.test.device import TestCase_CreateDevice, delete_device
from dtiot_d2c.d2c.test.testcli_base import TestCase, TestCLI_Base, TestSet
from dtiot_d2c.d2c.utils import color as color
from dtiot_d2c.dmo import ApiClient

log = logging.getLogger(__name__)
log_level = log.getEffectiveLevel() 

def delete_device_group(api:ApiClient, name:str):
    def operation():      
        if (application := DeviceGroup.get(api, name=name)):
            DeviceGroup.delete(api, name=name)        
    tutils.exec_test(f"Deleting device group {name}", operation)
    
class TestCase_CreateDeviceGroup(TestCase):
    def get_name(self):
        return "Create DeviceGroup"
    
    def run(self):
        name = self.cfg.get("name")

        create_props = self.cfg.get("create_props", None)
        if not create_props:
            raise Exception(f"Cannot find element create_props in configuration.")             

        def operation():   
            response = DeviceGroup.create(self.api, name=name, **create_props)

            if not response:
                return f"Device group {name} hasn't been created"
            
            if log_level <= tutils.INFO:
                GetResponse(response=response).print()      

        tutils.exec_test(f"Creating device group {name}", operation)

    def verify(self):
        name = self.cfg.get("name")

        create_props = self.cfg.get("create_props", None)
        if not create_props:
            raise Exception(f"Cannot find element create_props in configuration.")             
        
        '''
        {
            "cnd": "com.telekom.iot.orchestrator.deviceGroup",
            "rn": "myDeviceGroupFlexContainer000112",
            "nds": [
                "67d940fd665ccfe8ac0b736f"
            ],
            "aes": [
                "6724c9daa3618fe8cb390e24"
            ],
            "ty": 28,
            "cr": "CDevice-Group",
            "tsn": "dtiot:devGr",
            "st": 0,
            "ri": "67f3e2c1b59b8e3d0d98f0de",
            "pi": "67d85887d7ca00dc06b063b6",
            "ct": "20250407T143545,701000",
            "lt": "20250407T143545,701000",
            "m2m:sub": [
                {
                    "rn": "groupflexcontainer-update-subscription",
                    "nct": 1,
                    "enc": {
                        "net": [
                            1,
                            2,
                            3,
                            4
                        ]
                    },
                    "nu": [
                        "http://d2c-device-provisioning.d2c.svc.cluster.local/provisioning/device-group"
                    ],
                    "su": "http://d2c-device-provisioning.d2c.svc.cluster.local:80/provisioning/device-group",
                    "ty": 23,
                    "ri": "67f3e2c1136d98dce0994717",
                    "pi": "67f3e2c1b59b8e3d0d98f0de",
                    "ct": "20250407T143545,831000",
                    "lt": "20250407T143545,831000"
                }
            ]
        }
        '''
        def operation():
            onem2m_dict = dmo.get_resources(api_client=self.api, resourceName=f"device-group/{name}")
            if not onem2m_dict:
                raise Exception(f"Device group resource {name} couldn't be get after creation.")
            else:
                return onem2m_dict

        onem2m_dict:dict = tutils.exec_test(f"Getting ae {name}", operation)

        elem_path = "rn"
        elem_value = name
        def operation():
            if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
                raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
        tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)

        for (key, val) in create_props.get("labels", {}).items():
            elem_path = "lbl"
            elem_value = f"{key}:{val}"
            def operation():
                if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
                    raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
            tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)
            
class TestCase_ModifyDeviceGroup(TestCase):
    def get_name(self):
        return "Modify DeviceGroup"
    
    def run(self):
        name = self.cfg.get("name")
        modify_props = self.cfg.get("modify_props", None)
        if not modify_props:
            raise Exception(f"Cannot find element modify_props in configuration.")   

        def operation():      
            response = DeviceGroup.update(self.api, name=name, **modify_props)
            
            if not response:
                raise Exception(f"Device group {name} hasn't been modified")
        
            if log_level <= tutils.INFO:
                GetResponse(response=response).print()      

        tutils.exec_test(f"Modifing device group {name}", operation)

    def verify(self):
        name = self.cfg.get("name")
        modify_props = self.cfg.get("modify_props", None)
        if not modify_props:
            raise Exception(f"Cannot find element modify_props in configuration.")   

        def operation():
            onem2m_dict = dmo.get_resources(api_client=self.api, resourceName=f"device-group/{name}")
            if not onem2m_dict:
                raise Exception(f"Device group resource {name} couldn't be get after creation.")
            else:
                return onem2m_dict
        onem2m_dict:dict = tutils.exec_test(f"Getting device group {name}", operation)

        for (key, val) in modify_props.get("labels", {}).items():
            elem_path = "lbl"
            elem_value = f"{key}:{val}"
            def operation():
                if not tutils.call_has_element_value(onem2m_dict, elem_path, elem_value):
                    raise Exception(f"Value of element {elem_path} does not meet required value {elem_value}.")
            tutils.exec_test(f"Verifing {elem_path} for {elem_value}", operation)

class TestCase_DeleteDeviceGroup(TestCase):
    def get_name(self):
        return "Delete DeviceGroup"
    
    def run(self):
        delete_device_group(self.api, self.cfg.get("name"))

    def verify(self):
        name = self.cfg.get("name")

        def operation():
            if dmo.get_resources(api_client=self.api, resourceName=f"device-group/{name}"):
                raise Exception(f"Delete request didn't delete device group {name}.")
        tutils.exec_test(f"Checking if device group {name} has been deleted", operation)
                
class TestCase_AddApplicationsToDeviceGroup(TestCase):
    def __init__(self, api: ApiClient, device_group_cfg:dict, application_cfgs:List[dict], 
                 verify_device_cfgs:List[dict]=None):
        super().__init__(api=api, cfg=device_group_cfg)
        self.application_cfgs = application_cfgs
        self.verify_device_cfgs = verify_device_cfgs

    def get_name(self):
        return "'Add Applications to DeviceGroup"
    
    def run(self):
        device_group_name = self.cfg.get("name")
        self.application_names = []
                
        for application_cfg in self.application_cfgs:
            self.application_names.append(application_cfg.get("name"))
 
        for application_name in self.application_names:
            def operation():
                response = DeviceGroup.update_applications(self.api, 
                                                           name=device_group_name,
                                                           add_names=[application_name]
                                                          )
                if log_level <= tutils.INFO:
                    GetResponse(response=response).print()                 
                    
            tutils.exec_test(f"Adding application name {application_name} to device group {device_group_name}", operation)                
        
        return None

    def verify(self):
        '''
        {
            "cnd": "com.telekom.iot.orchestrator.deviceGroup",
            "rn": "devicegroup02-919",
            "lbl": [
                "deviceType:SDI People Counter"
            ],
            "nds": [],
            "aes": [
                "6802bf3474f6cb71f6fe8f1f"
            ],
            "ty": 28,
            "cr": "Cd2ccli",
            "tsn": "dtiot:devGr",
            "st": 1,
            "ri": "6802bf3074f6cb71f6fe8f1d",
            "pi": "680128d474f6cb71f6fe8e86",
            "ct": "20250418T210800,306000",
            "lt": "20250418T210811,046000",
            "m2m:sub": [
                {
                    "rn": "groupflexcontainer-update-subscription",
                    "nct": 1,
                    "enc": {
                        "net": [
                            1,
                            2,
                            3,
                            4
                        ]
                    },
                    "nu": [
                        "http://d2c-device-provisioning.d2c.svc.cluster.local/provisioning/device-group"
                    ],
                    "su": "http://d2c-device-provisioning.d2c.svc.cluster.local:80/provisioning/device-group",
                    "ty": 23,
                    "ri": "6802bf300aa198ff890ab5d4",
                    "pi": "6802bf3074f6cb71f6fe8f1d",
                    "ct": "20250418T210800,402000",
                    "lt": "20250418T210800,402000"
                }
            ]
        }
        '''
        device_group_name = self.cfg.get("name")
        
        ###
        # Verify if the application names have been added to the device group
        for application_name in self.application_names:
            def operation():
                onem2m_dict = dmo.get_resources(api_client=self.api, resourceName=f"device-group/{device_group_name}")
                if not onem2m_dict:
                    raise Exception(f"Could not load device group resource {device_group_name}.")
                else:
                    return onem2m_dict
            onem2m_dict:dict = tutils.exec_test(f"Getting device group {device_group_name}", operation)
                    
            def operation():
                if application_name not in onem2m_dict["aes"]:
                    raise Exception(f"Could not find application id {application_name} in device group.")
            tutils.exec_test(f"Verifying application id {application_name}", operation)

        ###
        # Verify if the application subscriptions have been added to the device communication
        # message received container
        if not self.verify_device_cfgs:
            return
        
        for device_cfg in self.verify_device_cfgs:
            device_name = device_cfg.get("name")

            # Name of subscription: <device-group-name>-<application-name>-<device-name>,
            for application_cfg in self.application_cfgs:
                application_name = application_cfg.get("name")
                
                n = f"device-communication/{device_name}/received-msg/{device_group_name}-{application_name}-{device_name}"
                def operation():
                    onem2m_dict = dmo.get_resources(api_client=self.api, resourceType="subscription", resourceName=n)
                    if not onem2m_dict:
                        raise Exception(f"Couldn't get subscription {n}.")
                    else:
                        return onem2m_dict
                onem2m_dict:dict = tutils.exec_test(f"Getting application subscription {n}", operation)           
        
class TestCase_RemoveApplicationsFromDeviceGroup(TestCase):
    def __init__(self, api: ApiClient, device_group_cfg:dict, 
                 application_cfgs:List[dict], verify_device_cfgs:List[dict]=None):
        super().__init__(api=api, cfg=device_group_cfg)
        self.application_cfgs = application_cfgs
        self.verify_device_cfgs = verify_device_cfgs
        self.application_names:List[str] = []
        
    def get_name(self):
        return "Remove Applications from DeviceGroup"
    
    def run(self):
        device_group_name = self.cfg.get("name")

        for application_cfg in self.application_cfgs:
            self.application_names.append(application_cfg.get("name"))
 
        def operation():
            response = DeviceGroup.update_applications(self.api, 
                                                       name=device_group_name,
                                                       remove_names=self.application_names
                                                      )
            if log_level <= tutils.INFO:
                GetResponse(response=response).print()                 
                
        tutils.exec_test(f"Removing application names {self.application_names} from device group {device_group_name}", operation)                
        
        return None

    def verify(self):
        '''
        {
            "cnd": "com.telekom.iot.orchestrator.deviceGroup",
            "rn": "devicegroup02-919",
            "lbl": [
                "deviceType:SDI People Counter"
            ],
            "nds": [],
            "aes": [
                "6802bf3474f6cb71f6fe8f1f"
            ],
            "ty": 28,
            "cr": "Cd2ccli",
            "tsn": "dtiot:devGr",
            "st": 1,
            "ri": "6802bf3074f6cb71f6fe8f1d",
            "pi": "680128d474f6cb71f6fe8e86",
            "ct": "20250418T210800,306000",
            "lt": "20250418T210811,046000",
            "m2m:sub": [
                {
                    "rn": "groupflexcontainer-update-subscription",
                    "nct": 1,
                    "enc": {
                        "net": [
                            1,
                            2,
                            3,
                            4
                        ]
                    },
                    "nu": [
                        "http://d2c-device-provisioning.d2c.svc.cluster.local/provisioning/device-group"
                    ],
                    "su": "http://d2c-device-provisioning.d2c.svc.cluster.local:80/provisioning/device-group",
                    "ty": 23,
                    "ri": "6802bf300aa198ff890ab5d4",
                    "pi": "6802bf3074f6cb71f6fe8f1d",
                    "ct": "20250418T210800,402000",
                    "lt": "20250418T210800,402000"
                }
            ]
        }
        '''
        name = self.cfg.get("name")
        
        def operation():
            onem2m_dict = dmo.get_resources(api_client=self.api, resourceName=f"device-group/{name}")
            if not onem2m_dict:
                raise Exception(f"Could not load device group resource {name}.")
            else:
                return onem2m_dict
        onem2m_dict:dict = tutils.exec_test(f"Getting device group {name}", operation)
                
        for application_name in self.application_names:
            def operation():
                if application_name in onem2m_dict["aes"]:
                    raise Exception(f"Application name {application_name} has NOT been removed from device group.")
            tutils.exec_test(f"Verifying application name {application_name}", operation)

        ###
        # Verify if the application subscriptions have been removed from the device communication
        # message received container
        if not self.verify_device_cfgs:
            return
        
        for device_cfg in self.verify_device_cfgs:
            device_name = device_cfg.get("name")

            # Name of subscription: <device-group-name>-<application-name>-<device-name>,
            for application_cfg in self.application_cfgs:
                application_name = application_cfg.get("name")
                
                n = f"device-communication/{device_name}/received-msg/{name}-{application_name}-{device_name}"
                def operation():
                    onem2m_dict = dmo.get_resources(api_client=self.api, resourceType="subscription", resourceName=n)
                    if onem2m_dict:
                        raise Exception(f"Application subscription {n} has not been removed.")
                tutils.exec_test(f"Verifying if application subscription {n} has been removed", operation)           

class TestCase_AddDevicesToDeviceGroup(TestCase):
    def __init__(self, api: ApiClient, device_group_cfg:dict, 
                 device_cfgs:List[dict], verify_application_cfgs:List[dict]=None):
        super().__init__(api=api, cfg=device_group_cfg)
        self.device_cfgs = device_cfgs
        self.verify_application_cfgs = verify_application_cfgs
        self.device_names = []
        
    def get_name(self):
        return "'Add Device to DeviceGroup"
    
    def run(self):
        device_group_name = self.cfg.get("name")

        for device_cfg in self.device_cfgs:
            self.device_names.append(device_cfg.get("name"))
 
        for device_name in self.device_names:
            def operation():
                response = DeviceGroup.update_devices(self.api, 
                                                       name=device_group_name,
                                                       add_names=[device_name]
                                                      )
                if log_level <= tutils.INFO:
                    GetResponse(response=response).print()                 
                    
            tutils.exec_test(f"Adding device name {device_name} to device group {device_group_name}", operation)                
        
        return None


    def verify(self):
        '''
        device-communication/<device-name>/
        {
            "rn": "device01-916",
            ....
            "m2m:cnt": [
                ...
                {
                    "mbs": 1000000,
                    "mni": 100,
                    "rn": "received-msg",
                    "ty": 3,
                    "st": 0,
                    "cbs": 0,
                    "cni": 0,
                    "cr": "CDevice-Provisioning",
                    "ri": "680b565083dba99abdafb5d3",
                    "pi": "680b565083dba99abdafb5cd",
                    "ct": "20250425T093056,295000",
                    "lt": "20250425T093056,295000",
                    "m2m:sub": [
                        {
                            "rn": "devicegroup02-916-devicegroup02-app1-916-device01-916",
                            "nct": 1,
                            "enc": {
                                "net": [
                                    1,
                                    2,
                                    3,
                                    4
                                ]
                            },
                            "nu": [
                                "d2c-dev-1/devicegroup02-app1-916"
                            ],
                            "su": "d2c-dev-1/devicegroup02-app1-916",
                            "ty": 23,
                            "ri": "680b566083dba99abdafb5df",
                            "pi": "680b565083dba99abdafb5d3",
                            "ct": "20250425T093112,448000",
                            "lt": "20250425T093112,448000"
                        }
                    ]
                },
                ...
            ]
        }        
        '''
        device_group_name = self.cfg.get("name")

        # Load device group
        def operation():
            onem2m_dict = dmo.get_resources(api_client=self.api, resourceName=f"device-group/{device_group_name}")
            if not onem2m_dict:
                raise Exception(f"Could not load device group resource {device_group_name}.")
            else:
                return onem2m_dict
        onem2m_dict:dict = tutils.exec_test(f"Getting device group {device_group_name}", operation)
          
        # Check if the device ids can be found in the device group   
        for device_name in self.device_names:
            def operation():
                if device_name not in onem2m_dict["nds"]:
                    raise Exception(f"Could not find device id {device_name} in device group.")
            tutils.exec_test(f"Verifying device id {device_name}", operation)
            
        ###
        # Check the application subscription in the received message device communication container
        # Name of subscription: <device-group-name>-<application-name>-<device-name>,
        if not self.verify_application_cfgs:
            return
        
        for device_cfg in self.device_cfgs:
            device_name = device_cfg.get("name")
            
            for application_cfg in self.verify_application_cfgs:
                application_name = application_cfg.get("name")
                
                n = f"device-communication/{device_name}/received-msg/{device_group_name}-{application_name}-{device_name}"
                def operation():
                    onem2m_dict = dmo.get_resources(api_client=self.api, resourceType="subscription", resourceName=n)
                    if not onem2m_dict:
                        raise Exception(f"Couldn't get subscription {n}.")
                    else:
                        return onem2m_dict
                onem2m_dict:dict = tutils.exec_test(f"Getting application subscription {n}", operation)   

class TestCase_RemoveDevicesFromDeviceGroup(TestCase):
    def __init__(self, api: ApiClient, device_group_cfg:dict, 
                 device_cfgs:List[dict], verify_application_cfgs:List[dict]=None):
        super().__init__(api=api, cfg=device_group_cfg)
        self.device_cfgs = device_cfgs
        self.verify_application_cfgs = verify_application_cfgs
        self.device_names:List[str] = []
        
    def get_name(self):
        return "Remove Devices from DeviceGroup"

    def run(self):
        device_group_name = self.cfg.get("name")

        for device_cfg in self.device_cfgs:
            self.device_names.append(device_cfg.get("name"))
 
        def operation():
            response = DeviceGroup.update_devices(self.api, 
                                                  name=device_group_name,
                                                  remove_names=self.device_names
                                                 )
            if log_level <= tutils.INFO:
                GetResponse(response=response).print()                 
                
        tutils.exec_test(f"Removing device names {self.device_names} from device group {device_group_name}", operation)   

    def verify(self):
        '''
        {
            "cnd": "com.telekom.iot.orchestrator.deviceGroup",
            "rn": "devicegroup02-919",
            "lbl": [
                "deviceType:SDI People Counter"
            ],
            "nds": [],
            "aes": [
                "6802bf3474f6cb71f6fe8f1f"
            ],
            "ty": 28,
            "cr": "Cd2ccli",
            "tsn": "dtiot:devGr",
            "st": 1,
            "ri": "6802bf3074f6cb71f6fe8f1d",
            "pi": "680128d474f6cb71f6fe8e86",
            "ct": "20250418T210800,306000",
            "lt": "20250418T210811,046000",
            "m2m:sub": [
                {
                    "rn": "groupflexcontainer-update-subscription",
                    "nct": 1,
                    "enc": {
                        "net": [
                            1,
                            2,
                            3,
                            4
                        ]
                    },
                    "nu": [
                        "http://d2c-device-provisioning.d2c.svc.cluster.local/provisioning/device-group"
                    ],
                    "su": "http://d2c-device-provisioning.d2c.svc.cluster.local:80/provisioning/device-group",
                    "ty": 23,
                    "ri": "6802bf300aa198ff890ab5d4",
                    "pi": "6802bf3074f6cb71f6fe8f1d",
                    "ct": "20250418T210800,402000",
                    "lt": "20250418T210800,402000"
                }
            ]
        }
        '''
        name = self.cfg.get("name")
        
        def operation():
            onem2m_dict = dmo.get_resources(api_client=self.api, resourceName=f"device-group/{name}")
            if not onem2m_dict:
                raise Exception(f"Could not load device group resource {name}.")
            else:
                return onem2m_dict
        onem2m_dict:dict = tutils.exec_test(f"Getting device group {name}", operation)
        
        for device_name in self.device_names:      
            def operation():
                if device_name in onem2m_dict["nds"]:
                    raise Exception(f"Device id {device_name} has NOT been removed from device group.")
            tutils.exec_test(f"Verifying device id {device_name}", operation)
            
        ###
        # Check the application subscription in the received message device communication container
        # Name of subscription: <device-group-name>-<application-name>-<device-name>,
        if not self.verify_application_cfgs:
            return
        
        for device_cfg in self.device_cfgs:
            device_name = device_cfg.get("name")
            
            for application_cfg in self.verify_application_cfgs:
                application_name = application_cfg.get("name")
                
                n = f"device-communication/{device_name}/received-msg/{name}-{application_name}-{device_name}"
                def operation():
                    onem2m_dict = dmo.get_resources(api_client=self.api, resourceType="subscription", resourceName=n)
                    if onem2m_dict:
                        raise Exception(f"Application subscription {n} still exists.")
                tutils.exec_test(f"Verifying if application subscription {n} has been deleted", operation)   
class TestCase_DeleteDeviceGroup(TestCase):
    def get_name(self):
        return "Delete DeviceGroup"
    
    def run(self):
        delete_device_group(self.api, self.cfg.get("name"))

    def verify(self):
        name = self.cfg.get("name")

        def operation():
            if dmo.get_resources(api_client=self.api, resourceName=f"device-group/{name}"):
                raise Exception(f"Delete request didn't delete device group {name}.")
        tutils.exec_test(f"Checking if device group {name} has been deleted", operation)
#MARK: Tests01
###########################################################################            
# Performs the following test cases:
# - create device group
# - modify device group
# - delete device group
###########################################################################    
class Tests01(TestSet):
    def __init__(self, api: ApiClient):
        super().__init__(api, "TestGroup Tests 01")
    
    def get_sample_configuration_varname(self)->str:
        return "DeviceGroup01Cfg"

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return [
            TestCase_CreateDeviceGroup(api=self.api, cfg=cfg),
            TestCase_ModifyDeviceGroup(api=self.api, cfg=cfg),
            TestCase_DeleteDeviceGroup(api=self.api, cfg=cfg),
        ]
        
    def test_pre_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))

    def test_post_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))

class Tests01_CLI(TestCLI_Base):
    def get1LineHelp(self):
        return "Runs device group test case: create, modify and delete device group."
    def get_test_set(self, api: ApiClient):
        return Tests01(api)
#MARK: Tests02
###########################################################################            
# Performs the following test cases:
# - create device group
# - create application
# - create application
# - remove application
# - add 2 applications 
# - remove application
# - remove application
###########################################################################    
class Tests02(TestSet):
    def __init__(self, api: ApiClient):
        super().__init__(api, "TestGroup Tests 02")
    
    def get_sample_configuration_varname(self)->str:
        return "DeviceGroup02Cfg"

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return [
            TestCase_CreateDeviceGroup(api=self.api, cfg=cfg),
            TestCase_CreateApplication(api=self.api, cfg=cfg.get("application_1")),
            TestCase_CreateApplication(api=self.api, cfg=cfg.get("application_2")),
            TestCase_AddApplicationsToDeviceGroup(api=self.api, 
                                                  device_group_cfg=cfg, 
                                                  application_cfgs=[cfg.get("application_1")]),
            TestCase_RemoveApplicationsFromDeviceGroup(api=self.api, 
                                                      device_group_cfg=cfg, 
                                                      application_cfgs=[cfg.get("application_1")]),
            TestCase_AddApplicationsToDeviceGroup(api=self.api, 
                                                  device_group_cfg=cfg, 
                                                  application_cfgs=[cfg.get("application_1"), 
                                                                    cfg.get("application_2")]),
            TestCase_RemoveApplicationsFromDeviceGroup(api=self.api, 
                                                      device_group_cfg=cfg, 
                                                      application_cfgs=[cfg.get("application_1")]),
            TestCase_RemoveApplicationsFromDeviceGroup(api=self.api, 
                                                      device_group_cfg=cfg, 
                                                      application_cfgs=[cfg.get("application_2")]),
        ]
        
    def test_pre_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))
        delete_application(self.api, cfg.get("application_1").get("name"))
        delete_application(self.api, cfg.get("application_2").get("name"))

    def test_post_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))
        delete_application(self.api, cfg.get("application_1").get("name"))
        delete_application(self.api, cfg.get("application_2").get("name"))

class Tests02_CLI(TestCLI_Base):
    def get1LineHelp(self):
        return "Runs device group test case: add applications, remove applications without devices."

    def get_test_set(self, api: ApiClient):
        return Tests02(api)
#MARK: Tests03
###########################################################################            
# Performs the following test cases:
# - create device group
# - create device
# - create device
# - add device to device group
# - remove device group
# - add 2 devices to device group
# - remove device group
# - remove device group
###########################################################################    
class Tests03(TestSet):
    def __init__(self, api: ApiClient):
        super().__init__(api, "TestGroup Tests 03")
    
    def get_sample_configuration_varname(self)->str:
        return "DeviceGroup02Cfg"

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return [
            TestCase_CreateDeviceGroup(api=self.api, cfg=cfg),
            
            TestCase_CreateDevice(api=self.api, cfg=cfg.get("device_1")),
            TestCase_CreateDevice(api=self.api, cfg=cfg.get("device_2")),
            TestCase_AddDevicesToDeviceGroup(api=self.api, 
                                            device_group_cfg=cfg, 
                                            device_cfgs=[cfg.get("device_1")],
                                            verify_application_cfgs=None),
            TestCase_RemoveDevicesFromDeviceGroup(api=self.api, 
                                                  device_group_cfg=cfg, 
                                                  device_cfgs=[cfg.get("device_1")]),
            TestCase_AddDevicesToDeviceGroup(api=self.api, 
                                            device_group_cfg=cfg, 
                                            device_cfgs=[cfg.get("device_1"), cfg.get("device_2")],
                                            verify_application_cfgs=None),
            TestCase_RemoveDevicesFromDeviceGroup(api=self.api, 
                                                  device_group_cfg=cfg, 
                                                  device_cfgs=[cfg.get("device_1"), cfg.get("device_2")])
        ]
        
    def test_pre_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))
        delete_device(self.api, cfg.get("device_1").get("name"))
        delete_device(self.api, cfg.get("device_2").get("name"))

    def test_post_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))
        delete_device(self.api, cfg.get("device_1").get("name"))
        delete_device(self.api, cfg.get("device_2").get("name"))
class Tests03_CLI(TestCLI_Base):
    def get1LineHelp(self):
        return "Runs device group test case: add devices, remove devices without applications."

    def get_test_set(self, api: ApiClient):
        return Tests03(api)
#MARK: Tests04
###########################################################################            
# Performs the following test cases:
# - create device group
# - create application 
# - create device
# - add applications to device group
# - add device to device group
# - remove device from device group
# - add device to device group
# - remove application from device group 
###########################################################################    
class Tests04(TestSet):
    def __init__(self, api: ApiClient):
        super().__init__(api, "TestGroup Tests 04")
    
    def get_sample_configuration_varname(self)->str:
        return "DeviceGroup02Cfg"

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return [
            TestCase_CreateDeviceGroup(api=self.api, cfg=cfg),
            TestCase_CreateApplication(api=self.api, cfg=cfg.get("application_1")),
            TestCase_CreateApplication(api=self.api, cfg=cfg.get("application_2")),
            TestCase_CreateDevice(api=self.api, cfg=cfg.get("device_1")),
            TestCase_CreateDevice(api=self.api, cfg=cfg.get("device_2")),
            TestCase_AddApplicationsToDeviceGroup(api=self.api, 
                                                  device_group_cfg=cfg, 
                                                  application_cfgs=[cfg.get("application_1"), cfg.get("application_2")],
                                                  verify_device_cfgs=None),
            TestCase_AddDevicesToDeviceGroup(api=self.api, 
                                            device_group_cfg=cfg, 
                                            device_cfgs=[cfg.get("device_1"), cfg.get("device_2")],
                                            verify_application_cfgs=[cfg.get("application_1"), cfg.get("application_2")]),
            TestCase_RemoveDevicesFromDeviceGroup(api=self.api, 
                                                  device_group_cfg=cfg, 
                                                  device_cfgs=[cfg.get("device_1")],
                                                  verify_application_cfgs=[cfg.get("application_1")]),
            TestCase_RemoveApplicationsFromDeviceGroup(api=self.api, 
                                                      device_group_cfg=cfg, 
                                                      application_cfgs=[cfg.get("application_1")],
                                                      verify_device_cfgs=[cfg.get("device_1")]),
        ]
        
    def test_pre_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))
        delete_application(self.api, cfg.get("application_1").get("name"))
        delete_device(self.api, cfg.get("device_1").get("name"))

    def test_post_cleanup(self, cfg:dict)->str:
        delete_device_group(self.api, cfg.get("name"))
        delete_application(self.api, cfg.get("application_1").get("name"))
        delete_device(self.api, cfg.get("device_1").get("name"))

class Tests04_CLI(TestCLI_Base):
    def get1LineHelp(self):
        return "Runs device group test case: add and remove application and device."

    def get_test_set(self, api: ApiClient):
        return Tests04(api)    



import importlib
import json
import logging
import os
import os.path
import sys
from datetime import datetime
from pathlib import Path

log = logging.getLogger(__name__)

class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'

def splitString(str, sepChar):
    a = []
    for s in str.split(sepChar):
        a.append(s.strip())
    return a

def isDict(obj):
    if not obj:
        return False
    else:
        return True if obj.__class__ == {}.__class__ else False

def isArray(obj):
    if not obj:
        return False
    else:
        return True if obj.__class__ == [].__class__ else False

def isString(obj):
    return obj.__class__ == "str".__class__ if obj else False

def convert_str_to_python_type(s:str):
    if not s:
        return s
    
    try:
        return int(s)
    except Exception:
        pass
    
    try:
        return float(s)
    except Exception:
        pass
    
    if s.lower() == 'true':
        return True
    elif s.lower() == 'false':
        return False

    try:
        result = json.loads(s)
        if isinstance(result, (list, dict, tuple)):
            return result
    except json.JSONDecodeError:
        pass            
    
    return s

from enum import Enum


class ReturnType(Enum):
    TEXT = 1
    JSON_AS_DICT = 2
    LINES_AS_LIST = 3

def execCmd(cmd, returnType : ReturnType = ReturnType.TEXT):
    log.debug("Exec %s" % (cmd))
        
    cmdIn = os.popen(cmd)

    line = cmdIn.readline()
    s = ""
    while line:
        if line.startswith("b\'"):
            line = line.strip()
            line = line[2:-1] + "\n"
        s = s + line
        line = cmdIn.readline()    

    log.debug("\n%s" % (s.strip()))                                  

    if returnType == ReturnType.JSON_AS_DICT:
        return json.loads(s)
    elif returnType == ReturnType.LINES_AS_LIST:
        return splitString(s, "\n")
    else:
        return s
    
class FileLinesIterator:
    '''
    Iterator class to walk through lines of a text file in the context of for loop.

    Sample code:    

    from file_lines_iter import FileLines

    for line in FileLinesIterator("/Users/rolan/Desktop/WORK/1646599091.json"):
        print(line)
        
    '''
    
    def __init__(self, filepath, stripLines:bool=False, skipEmptyLines:bool=False, skipLinesStartWith:list[str]=[]):
        '''
        Constructor of FileLinesIterator class.
        :param filepath: Path of the file through which lines shall be iterated. If "-" than stdin is read
        :param skipLines: If set all lines are left and right stripped.
        :param skipEmptyLines: If set to true empty lines are not returned.
        :param skipLinesStartWith: Array of single characters. If a line starts with one of the chars it is not returned.
        '''
        self._filepath = filepath
        self._file = None
        self._skipEmptyLines=skipEmptyLines
        self._skipLinesStartWith=skipLinesStartWith
        self._stripLines=stripLines

    def __iter__(self):
        return self
    
    def __next__(self):
        if not self._file:
            if self._filepath=="-":
                self._file = sys.stdin
            else:
                self._file = open(self._filepath, "r")

        while True:
            line = self._file.readline()
            if not line:
                if self._filepath!="-":
                    self._file.close()
                raise StopIteration

            if self._stripLines:
                line=line.strip()
            else:
                line=line.rstrip()
                
            if self._skipEmptyLines and len(line)==0:
                continue

            if len(line)>0 and line[0] in self._skipLinesStartWith:
                continue

            return line

import socket


def getIp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

# Class to executues a function every x seconds 
import threading
import time
import traceback


class FunctionExecutor(threading.Thread):
    def __init__(self, cronSecs, function, *args, **kwargs):
        super().__init__()
        self._cronSecs = cronSecs
        self._function = function
        self._args = args
        self._kwargs = kwargs

    def run(self):
        while True:
            try:
                time.sleep(self._cronSecs)
                self._function(*self._args, **self._kwargs)
            except Exception as ex:
                log.error(str(ex))
                traceback.print_exc()
                
def getValueFromDict(rootD:dict, elemPath:str, default:any=None, sepChar=".") -> any:
    '''
    Gets a value from a dictionary element.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    name = names[0]    

    if name not in d.keys() or len(names) == 0:
        return default
    elif len(names) == 1:
        return d.get(name)            
    else:
        return getValueFromDict(d.get(names[0]), sepChar.join(names[1:]), default=default, sepChar=sepChar)

def hasElementInDict(rootD:dict, elemPath:str, sepChar=".") -> bool:
    '''
    Verifies if a specific element in the dictionary exists.
    Returns True if the element exists. 
    Return False if the element not exists.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    firstName = names[0]    

    if firstName not in d.keys() or len(names) == 0:
        return False
    elif firstName not in d.keys():
        return False
    elif len(names) == 1:
        return True if firstName in d.keys() else False          
    else:
        return hasElementInDict(d.get(names[0]), sepChar.join(names[1:]), sepChar=sepChar)

def setValueInDict(rootD:dict, elemPath:str, value:any, sepChar=".", createIfNotExist=False) -> bool:
    '''
    Sets a value to an element in a dictionary.
    Returns True if the value of the element has been set.
    Return False if the value hasn't been set because the element does not exist.
    '''
    d = rootD
    names = elemPath.split(sepChar)
    firstName = names[0]    

    if firstName not in d.keys():
        if createIfNotExist:
            if len(names) > 1:
                for i in range(0, len(names) - 1):
                    d[names[i]] = {}
                    d = d[names[i]]
            d[names[-1]] = value
            return True
        else:
            return False

    # At this point the element with firstName always exist.
    # If we reached the leaf of the path
    if len(names) == 1:
        d[firstName] = value
        return True
    else:
        return setValueInDict(d.get(names[0]), sepChar.join(names[1:]), value, sepChar=sepChar, createIfNotExist=createIfNotExist)

def flattenDict(inD:dict, prefix:str="", sepChar:str="_")->dict:
    '''
    Walks top/down through a multi-level dictionary and creates out of it a dictionary
    with only a single level. The names of the sub-dictionary in the input dictionary
    are separated in the name of the return dictionary by a separator character.
    
    :param inD:dict:     Mulit level input dictionary which shall be converted into a single level dictionary.
    :param prefix:str:   Name prefix which shall be used.
    :sepChar:str:        Seperator character.A
    
    :return:dict:       Single level dictionary
    '''
    retD = {}
    for k in inD.keys():
        e = inD[k]
        if type(e) == dict:
            d = flattenDict(e, prefix=f"{prefix}{k}{sepChar}", sepChar=sepChar)
            retD.update(d)
        else:
            retD[f"{prefix}{k}"] = inD[k]

    return retD

def get_value_from_dict_3(rootD:dict, elemPath:str, default:any=None, sepChar=".") -> any:
    # If rootD is not a dictionary but a list, call this function for each element 
    # of the list
    if type(rootD) in [list, tuple]:
        values = []
        for e in rootD:
            v = get_value_from_dict_3(e, elemPath, default=default, sepChar=sepChar)
            values.append(v)
        return values
    elif type(rootD) != dict:
        return ""
    
    d = rootD
    names = elemPath.split(sepChar)
    name = names[0]    

    # If name represents an elemement in an array such as dvi[0]
    list_idx = -1
    
    if (idx1 := name.find("[")) > -1:
        if (idx2 := name.find("]", idx1)) == -1:
            return default
        list_idx = int(name[idx1+1:idx2])
        name = name[0:idx1]
        
    if name not in d.keys() or len(names) == 0:
        return default
    elif len(names) == 1:
        if list_idx == -1:
            return d.get(name)            
        else:
            l = d.get(name)
            return l[list_idx] if list_idx < len(l) else default
    else:
        if list_idx == -1:
            return get_value_from_dict_3(d.get(names[0]), sepChar.join(names[1:]), default=default, sepChar=sepChar)
        else:
            l = d.get(name)
            if list_idx >= len(l):
                return default
            return get_value_from_dict_3(l[list_idx], sepChar.join(names[1:]), default=default, sepChar=sepChar)
    
def suppressEx(func, *args, **kwargs) -> any:
    '''
    Runs a function and suppresses an exception. 
    
    :param *args: Optional arguments to pass into the function.
    :param **kwargs: Option keyword arguments to pass into the function.
    
    :return: Return of the function or None in case an exception is caught.
    '''
    try:
        return func(*args, **kwargs)
    except Exception as ex:
        log.error(f"Error while exec {func.__name__}: {ex}")        
        return None           
    
########################################################################
#MARK: base64
########################################################################
import base64


def str_to_b64str(s:str)->str:
    if not s:
        return s
    b = s.encode('utf-8')
    b64 = base64.b64encode(b)
    return b64.decode('utf-8')    
    
def to_basic_auth(username:str, password:str)->str:
    return str_to_b64str(f"{username}:{password}")

###
# Some usefull decorators
#MARK: profile decorator
from functools import wraps


def profile(log_func=None,       # Logger function to use for output. 
                                 # If not defined "print" is used to output to stdout.
            log_prefix="profile" # Prefix string for the log output
):                               # Any value of decorated function
    '''
    Function decorator to print function call arguments and execution time to stdout or a logger.
    
    By default the profile decoration is switchd off. To switch it on export the environment 
    variable DTIOT_PROFILE=1.
    
    How to use the profile decorator:

    import utils

    @utils.profile(log_func=log.info)
    def myFunc(name):
        print(f"Hallo from {name}")
    '''
    
    if not log_func:
        log_func = print
        
    def decorator(fu):
        if os.getenv("DTIOT_PROFILE", "false").lower() not in ["1", "true", "yes"]:
            return fu
        @wraps(fu)
        def wrapper(*args, **kwargs):
            s = ""
            for a in args:
                if s:
                    s += ", "
                s += f"{a}" 
            for (k, v) in kwargs.items():
                if s:
                    s += ", "
                s += f"{k}={v}" 
            log_func(f"{log_prefix}: calling {fu.__name__}({s})")
                
            before = time.time()
            rv = fu(*args, **kwargs)
            after = time.time()
            
            log_func(f"{log_prefix}: called {fu.__name__}: {after - before:.6f} secs")

            return rv
        return wrapper
    return decorator   
    
########################################################################
#MARK: date function
########################################################################
from datetime import datetime


def fnow(format:str="%Y-%m-%d %H:%M:%S")->str:
    return datetime.now().strftime(format)

class DatetimeJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        else:
            return super().default(obj)              
        
        


import argparse
import json
import logging
from argparse import ArgumentParser

from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.dmo import device_provisioning as device_provisioning
from dtiot_d2c.dmo import dmo as dmo
from dtiot_d2c.dmo import utils as utils

log = logging.getLogger(__name__)

class _CFG:
    default_d2c_origin:str = "NI-IPE"
                                      
cmdArgTemplates:dict = {
    "-lbl" : ["--labels", 
        {
            "metavar":"{key:value, key:value, ...}", 
            "dest":"labels", 
            "required":False,
            "default":None,
            "help":"Labels defined by a JSON dictionary of key/value pairs."
        }
    ],
    "-rn": ["--resource-name",
            {"metavar":"<string>", 
            "dest":"rn", 
            "required": False,
            "default":None,
            "help":"Name of the resource object."}
    ],
    "-ri": ["--resource-id", 
            {"metavar":"<string>", 
            "dest":"ri", 
            "required": False,
            "default":None,
            "help":"Id of the object."}
    ],        
    "-rt": ["--resource-type", 
            {"metavar":"<string>",
            "dest": "ty",
            "choices": dmo.TYs.keys(),
            "help":f"Type of resource. Following values are possible: {dmo.TYs.keys()}"}
    ],
    "--outfile": {"metavar":"<filepath>", 
            "dest":"outfile", 
            "required": False,
            "default":None,
            "help":"File into which the log outputs shall be written additionaly."
    },
    "-o": ["--origin", 
           {"metavar":"<string>", 
            "dest":"origin", 
            "required": False,
            "default":None,
            "help":"X-M2M-ORIGIN heder."}
    ],
    "-af": ["--attribute-filter", 
            {"metavar":"<string>", 
            "dest":"attr_filters", 
            "required": False,
            "default":None,
            "help":"Comma separated list of attribute filters such as: reqRD=67dc2219c9a47b53f222658b,st=0"}
    ],            
    "-rci": ["--result-content-indicator", 
            {"dest":"rci", 
            "choices": dmo.RCIs.keys(),
            "required": False,
            "default":"ac",
            "help":"Result content indicator. Default is attributes-and-children."}
    ],          
    "-li": ["--limit",
            {"metavar":"<int>", 
            "type":int,
            "dest":"limit", 
            "required": False,
            "default":100000000,
            "help":"Max. number objects to return."}
    ],             
    "-os": ["--offset",
            {"metavar":"<int>", 
            "type":int,
            "dest":"offset", 
            "required": False,
            "default":0,
            "help":"Offset for pagination retrievel."}
    ],             
    "-s": ["--select",
            {"metavar":"<select>", 
            "dest":"select", 
            "required": False,
            "help":"Comma separated list of elements to be selected."}
    ],             
    "-f": ["--format",
            {"dest":"format", 
            "choices": ["CSV", "JSON"], 
            "required": False,
            "help":"Output format for selected fields."}
    ],             
    "-sc": ["--separator-char",
            {"metavar":"<char>", 
            "dest":"sepchar", 
            "default":";",
            "required": False,
            "help":"Character to separate the fields in the select response. Default value is ;"}
    ],             
}        

# Definiere die Funktion für den benutzerdefinierten Typ
def dict_arg(s):
    try:
        return json.loads(s)
    except Exception as ex:
        raise argparse.ArgumentTypeError(f"{s} is not a valid JSON: {ex}")

def dict_file_arg(filepath):
    s:str = None
    try:
            f = open(filepath, "r")
            s =f.read()
            f.close()        
    except Exception as ex:
        raise argparse.ArgumentTypeError(f"Could not read file {filepath}: {ex}")

    try:
        return json.loads(s)
    except Exception as ex:
        raise argparse.ArgumentTypeError(f"{s} is not a valid JSON: {ex}")
#####################################################################################
class GetResourcesCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns resources."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-ri")        
        
        if not self.hasHideCmdArg("-rt"):        
            self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rt", required=False)

        if not self.hasHideCmdArg("-rci"):    
            argsParser.add_argument("-rci", "--result-content-indicator", dest="rci",
                                    choices = dmo.RCIs.keys(),
                                    default="ac",
                                    help="Result content indicator. Default is attributes-and-children.")

        if not self.hasHideCmdArg("-li"):    
            argsParser.add_argument("-li", "--limit", metavar='<int>', type=int, dest="limit", default=1000000000,
                                help="Max. number objects to return.")

        if not self.hasHideCmdArg("-os"):    
            argsParser.add_argument("-os", "--offset", metavar='<int>', type=int, dest="offset", default=0,
                                help="Index of the first object to return.")


        argsParser.add_argument("-s", "--select", metavar='<select>', dest="select",
                               help="Comma separated list of elements to be selected.")

        argsParser.add_argument("-f", "--format",
                                dest="format",
                                choices = ["CSV", "JSON"], 
                                help="Output format for selected fields.")

        argsParser.add_argument("-sc", "--separator-char", metavar='<char>', dest="sepchar", 
                                default=";",
                                help="Character to separate the fields in the select response. Default value is ;")
                
    def main(self, cmdargs=None):

        obj = dmo.get_resources(self.apiConn, 
                                origin=cmdargs.origin or self.apiConn.origin,
                                resourceType=cmdargs.ty,
                                resourceName=cmdargs.rn,
                                resourceId=cmdargs.ri,
                                resultContentIndocator=cmdargs.rci, 
                                limit=cmdargs.limit, 
                                offset=cmdargs.offset,
                                select=cmdargs.select,
                                format=cmdargs.format)

        if not cmdargs.select or cmdargs.format == "JSON":
            print(json.dumps(obj, sort_keys=False, indent=4))
        else:
            for a in obj:
                line = cmdargs.sepchar.join(a)
                print("%s" % (line))

#####################################################################################
class AddResourceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds an object of a specific resource type."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rt", required=True)
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn", required=False)
        
        argsParser.add_argument("-b", "--body", metavar="<JSON>", type=dict_arg, dest="body", 
                                default=None, required=False, 
                                help="OneM2M JSON body of the resource object to create.")

        argsParser.add_argument("-bf", "--body-file", metavar="<file>", type=dict_file_arg, dest="file_body",
                                default=False, required=False, 
                                help="File with the OneM2M JSON body.")
        
    def main(self, cmdargs=None):
        body = {}

        if cmdargs.body:
            body = cmdargs.body
        elif cmdargs.file_body:
            body = cmdargs.file_body

        # If an application entity shall be created and no origin has been defined
        # Use the name of the application entity as the origin. 
        # If the default origin "d2ccli" it is not possible to create application  
        # entities
        if cmdargs.ty in ("ae", "application-entity") and not cmdargs.origin:
            cmdargs.origin = body.get("m2m:ae", {}).get("rn", None)
    
        obj = dmo.add_resource(self.apiConn,
                               origin=cmdargs.origin,
                               r_type=cmdargs.ty, 
                               r_name=cmdargs.rn,
                               onem2m_body=body,
                              )

        print(json.dumps(obj, sort_keys=False, indent=4))
                  
#####################################################################################
class UpdateResourceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Updates a resource object."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-ri", required=False)
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn", required=False)
        
        argsParser.add_argument("-b", "--body", metavar="<JSON>", type=dict_arg, dest="body", 
                                default=None, required=False, 
                                help="OneM2M JSON body of the resource object to create.")

        argsParser.add_argument("-bf", "--body-file", metavar="<file>", type=dict_file_arg, dest="file_body",
                                default=False, required=False, 
                                help="File with the OneM2M JSON body.")
        
    def main(self, cmdargs=None):
        body = {}

        if cmdargs.body:
            body = cmdargs.body
        elif cmdargs.file_body:
            body = cmdargs.file_body
    
        obj = dmo.update_resource(self.apiConn,
                                  origin=cmdargs.origin,
                                  name=cmdargs.rn, 
                                  id=cmdargs.ri, 
                                  onem2m_body=body,
                                 )

        print(json.dumps(obj, sort_keys=False, indent=4))
                                                   
#####################################################################################
class GetDeviceProvisioningRequestsCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the device provisioning requests."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn")
        #self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-ri")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-af")
        
        argsParser.add_argument("-s", "--select", metavar='<select>', dest="select",
                               help="Comma separated list of elements to be selected.")

        argsParser.add_argument("-f", "--format",
                                dest="format",
                                choices = ["CSV", "JSON"], 
                                help="Output format for selected fields.")

        argsParser.add_argument("-sc", "--separator-char", metavar='<char>', dest="sepchar", 
                                default=";",
                                help="Character to separate the fields in the select response. Default value is ;")
                
    def main(self, cmdargs=None):

        if cmdargs.attr_filters:
            af = utils.splitString(cmdargs.attr_filters, ",")
        else:
            af = None
            
        obj = device_provisioning.get_requests(self.apiConn, 
                                               origin=cmdargs.origin or self.apiConn.origin,
                                               name=cmdargs.rn,
                                               #id=cmdargs.ri,
                                               select=cmdargs.select,
                                               format=cmdargs.format,
                                               attribute_filters=af)

        if not cmdargs.select or cmdargs.format == "JSON":
            print(json.dumps(obj, sort_keys=False, indent=4))
        else:
            for a in obj:
                line = cmdargs.sepchar.join(a)
                print("%s" % (line))        
                
#####################################################################################
class GetDeviceProvisioningResponsesCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the device provisioning requests."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-af")
        
        argsParser.add_argument("-s", "--select", metavar='<select>', dest="select",
                               help="Comma separated list of elements to be selected.")

        argsParser.add_argument("-f", "--format",
                                dest="format",
                                choices = ["CSV", "JSON"], 
                                help="Output format for selected fields.")

        argsParser.add_argument("-sc", "--separator-char", metavar='<char>', dest="sepchar", 
                                default=";",
                                help="Character to separate the fields in the select response. Default value is ;")
                
    def main(self, cmdargs=None):

        if cmdargs.attr_filters:
            af = utils.splitString(cmdargs.attr_filters, ",")
        else:
            af = None
            
        obj = device_provisioning.get_responses(self.apiConn, 
                                    origin=cmdargs.origin or self.apiConn.origin,
                                    name=cmdargs.rn,
                                    select=cmdargs.select,
                                    format=cmdargs.format,
                                    attribute_filters=af)

        if not cmdargs.select or cmdargs.format == "JSON":
            print(json.dumps(obj, sort_keys=False, indent=4))
        else:
            for a in obj:
                line = cmdargs.sepchar.join(a)
                print("%s" % (line))                     

class DeleteResourceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Deletes a resource object."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn", required=False)
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-ri", required=False)
                
    def main(self, cmdargs=None):
        apiConn = self.apiConn
                
        if not cmdargs.rn and not cmdargs.ri:
            raise Exception(f"--resource-name or --resource-id option required to identify the object.")
        
        obj = dmo.delete_resource(apiConn, origin=cmdargs.origin, name=cmdargs.rn, id=cmdargs.ri)

        if obj:
            print(str(obj))

                                    
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
#####################################################################################
class oldUpdateResourceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Updates an resource object."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rt", required=False)
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn", "--resource-name")
                
    def main(self, cmdargs=None):
        '''
        Main entry point.
        '''
        apiConn = self.apiConn

        obj = dmo.updateResource(apiConn, 
                                 resourceType=cmdargs.ty, 
                                 resourceName=cmdargs.rn,
                                 labels=cmdargs.labels)

        print(json.dumps(obj, sort_keys=False, indent=4))                                

#####################################################################################
class oldAddMessageContainerCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds a message container to an resource object."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn", "--resource-name")

        argsParser.add_argument("-cn", "--container-name", metavar="<string>", dest="containerName",
                                required=True,
                                help="Name of the container within the resource object. E.g. uplinkMsg or downlinkMsg")
        
        argsParser.add_argument("-mms", "--max-message-size", metavar="<bytes>", dest="maxByteSize",
                                required=False, type=int, default=10000, 
                                help="Max. size of a message. The default is 10.000 bytes.")

        argsParser.add_argument("-mmn", "--max-message-num", metavar="<num>", dest="maxNrOfInstaces",
                                required=False, type=int, default=100, 
                                help="Max. number of stored messages. The default is 100.")

        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-lbl", "--labels")
                
    def main(self, cmdargs=None):
        '''
        Main entry point.
        '''
        apiConn = self.apiConn
            
        obj = dmo.addMessageContainer(apiConn, 
                                     resourceName=cmdargs.rn,
                                     containerName=cmdargs.containerName,
                                     maxByteSize=cmdargs.maxByteSize,
                                     maxNrOfInstaces=cmdargs.maxNrOfInstaces,
                                     labels=cmdargs.labels)

        print(json.dumps(obj, sort_keys=False, indent=4))                 
        
#####################################################################################
class oldAddSubscriptionCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds a subscription to an resource object."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-rn", "--resource-name")

        argsParser.add_argument("-sn", "--subscription-name", metavar="<string>", dest="subscriptionName",
                                required=True,
                                help="Name of the subscription  within the resource object.")
        
        argsParser.add_argument("-nu", "--notification-url", metavar="<url>", dest="notificationUrl",
                                required=True,
                                help="URL to which the notification shall be forwarded.")

        argsParser.add_argument("-nct", "--notification-content-type", dest="notificationContentType",
                                choices = dmo.NCTs.keys(), default="all",
                                help="Content type of the notification event. Defaule is all")

        argsParser.add_argument("-enc", "--event-notification-criteria", dest="eventNotificationCriteria",
                                choices=dmo.ENCs.keys(), default="update,delete,create-child,delete-child", 
                                metavar="[<string>,<string>,...]",
                                help=f"Comma separated list of event notification criterias. Default is u,d,cc,dc. Possible values are {dmo.ENCs.keys()}")

        self.addCmdLineArgsFromTemplate(cmdArgTemplates, argsParser, "-lbl", "--labels")
                
    def main(self, cmdargs=None):
        '''
        Main entry point.
        '''
        apiConn = self.apiConn

        encs = []
        if cmdargs.eventNotificationCriteria:
            for s in utils.splitString(cmdargs.eventNotificationCriteria,","):
                encs.append(s)

                            
        obj = dmo.addSubscription(apiConn, 
                                  resourceName=cmdargs.rn,
                                  subscriptionName=cmdargs.subscriptionName,
                                  notificationUrl=cmdargs.notificationUrl,
                                  notificationContentType=cmdargs.notificationContentType,
                                  eventNotificationCriterias=encs,
                                  pendingNotification="sendAllPending",
                                  labels=cmdargs.labels
                                  )
                    
        print(json.dumps(obj, sort_keys=False, indent=4))                 
        



import json
import logging
import os
import sys
import time

import dtiot_d2c.d2c.test.utils as tutils
import dtiot_d2c.dmo as dmo
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.application import Application
from dtiot_d2c.d2c.device.consts import *
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.dmo import ApiClient
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
            
def print_flushed(s:str, end=None):
    print(s, end=end)
    sys.stdout.flush() 

class PrintSampleTestConfigurationCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()
        
    def get1LineHelp(self):
        return "Prints a sample configuration for all test cases."
        
    def addCmdLineArgs(self, argsParser):
        argsParser.add_argument("-co", "--config-outfile", metavar="<file>", 
                                dest="out_filepath", required=False,
                                help="File into which the sample test configuration shall be written to. If not defined it is writtten to stdout.")

    def main(self, cmdargs=None):

        config_filepath = f"{os.path.dirname(__file__)}/sample_test_configuration.py"
        in_file = open(config_filepath, "r")
        
        if not cmdargs.out_filepath:
            print(in_file.read())
        else:
            out_file = open(cmdargs.out_filepath, "w")     
            out_file.write(in_file.read())
            out_file.close()



import json
import logging
from argparse import ArgumentParser

import dtiot_d2c.d2c as d2c
import dtiot_d2c.dmo as dmo
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.application import Application
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
          
#####################################################################################                
class AddC2cCliApplicationCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Creates the application for the d2c command line interface it-self."
    
    def addCmdLineArgs(self, argsParser):
        pass
                
    def main(self, cmdargs=None):
        response = Application.create(self.apiConn, 
                                      origin=self.apiConn.origin,
                                      name=self.apiConn.origin,
                                     )

        GetResponse(response=response).print()      
                    
#####################################################################################                
class AddApplicationCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Creates a new application."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n", required=True)
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-at")  # --application-type
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de")  # --description
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-us")  # --urls
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cp")  # --connection-properties
                
    def main(self, cmdargs=None):
        response = Application.create(self.apiConn, 
                                      origin=cmdargs.origin if cmdargs.origin else cmdargs.name,
                                      name=cmdargs.name,
                                      urls=cmdargs.urls,
                                      application_type=cmdargs.application_type,
                                      description=cmdargs.description,
                                      connection_properties=cmdargs.connection_properties
                                     )
        GetResponse(response=response).print()        
        
#####################################################################################
class GetApplicationsCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the applications."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o") #, default=dmo.CLICFG.default_d2c_origin)
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n") # --name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i") # --id
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-li")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-os")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-rci")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-s")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-f")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-sc")
                
    def main(self, cmdargs=None):
        response = Application.get(self.apiConn,
                                   origin=cmdargs.origin,
                                   id=cmdargs.id,
                                   name=cmdargs.name,
                                   limit=cmdargs.limit,
                                   offset=cmdargs.offset,
                                   resultContentIndocator=cmdargs.rci, 
                                   select=cmdargs.select,
                                   format=cmdargs.format)

        GetResponse(response=response).print(cmdargs.sepchar)
            
#####################################################################################
class UpdateApplicationCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Updates an application."

    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o") 
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de")  # --description
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-us")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cp")  # --connection-properties
                
    def main(self, cmdargs=None):
        response = Application.update(self.apiConn, 
                                      origin=cmdargs.origin or self.apiConn.origin,
                                      id=cmdargs.id,
                                      name=cmdargs.name,
                                      urls=cmdargs.urls,
                                      connection_properties=cmdargs.connection_properties,
                                      description=cmdargs.description,
                                     )
        if not response:
            raise Exception(f"Application {cmdargs.name or cmdargs.id} does not exist")
                        
        GetResponse(response=response).print()   

#####################################################################################            
class DeleteApplicationCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Deletes an application."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
                
    def main(self, cmdargs=None):
        obj = Application.delete(self.apiConn, 
                                 origin=cmdargs.origin, 
                                 name=cmdargs.name, 
                                 id=cmdargs.id
                                )

        if obj:
            print(str(obj))                                    
            



import json
import os

NAME="DEFAULT"

LogLevel = os.getenv("LOGLEVEL") or "INFO" 
ServiceName = os.getenv("SERVICE_NAME") or "d2c-profile-management-api"
ServiceUrlPrefix = f"/{ServiceName}"
ServicePort = os.getenv("SERVICE_PORT") or 38081
AppVersion = os.getenv("APP_VERSION") or "X.X.X"      
ServiceHostname = os.getenv("SERVICEHOSTNAME") or "127.0.0.1"

if ServiceHostname == "127.0.0.1":
    ExternalUrl = f"http://{ServiceHostname}:{ServicePort}/{ServiceName}"
else:
    ExternalUrl = f"https://{ServiceHostname}/{ServiceName}"
    
WelcomeMessage = f"GET {ExternalUrl}/healthz to test"

DocsCfg = {
    "url_token" : os.getenv("DOCS_URL_TOKEN") or ""
}
DocsUrl = f'/docs'   




import sys
import os
import importlib.util
    
def _importModule(moduleName, moduleFile):
    spec = importlib.util.spec_from_file_location(moduleName, moduleFile)
    module = importlib.util.module_from_spec(spec)
    spec = spec.loader.exec_module(module)
    return module

def _loadConfigFile(cfgsDir, cfgName):
    cfgFile = "%s/%s.py" % (cfgsDir, cfgName)    

    # Import all attributes from the configuration file into the scope
    # of this module.
    if os.path.exists(cfgFile):
        cfgModule = _importModule(f"__cfg{cfgName}", cfgFile)
        thisModule = sys.modules[__name__]
        for k in dir(cfgModule):
            if not k.startswith("__"):
                setattr(thisModule, k, getattr(cfgModule, k))
    else:
        print("Cannot load configuration. Configuration variable ", cfgFile, " does not exist.", file=sys.stderr)
        
_tmp_ENVVARNAME_CFG="CFG"
_tmp_ENVVARNAME_CFGSDIR="CFGSDIR"
_tmp_CONST_CFGSDIRNAME="configs"

_tmp_cfgsDir = os.environ.get(_tmp_ENVVARNAME_CFGSDIR)
if not _tmp_cfgsDir:
    _tmp_m = sys.modules["__main__"]
    _tmp_file = _tmp_m.__file__
    if os.path.islink(_tmp_file):
        _tmp_file = os.path.realpath(_tmp_file)
    _tmp_cfgsDir = "%s/%s" % (os.path.dirname(_tmp_file), _tmp_CONST_CFGSDIRNAME)

# Load default configuration
_loadConfigFile(_tmp_cfgsDir, "DEFAULT")

# Overwrite the DEFAULTs with the environ specifics
_tmp_cfgName = os.environ.get(_tmp_ENVVARNAME_CFG)

if _tmp_cfgName:
    _loadConfigFile(_tmp_cfgsDir, _tmp_cfgName)
else:
    print("Cannot load configuration. Environment variable ", _tmp_ENVVARNAME_CFG, " not defined.", file=sys.stderr)

# if _tmp_cfgName:
#     _tmp_cfgFile = "%s/%s.py" % (_tmp_cfgsDir, _tmp_cfgName)    

#     # Import all attributes from the configuration file into the scope
#     # of this module.
#     if os.path.exists(_tmp_cfgFile):
#         _tmp_cfgModule = _importModule("__cfg", _tmp_cfgFile)
#         _tmp_thisModule = sys.modules[__name__]
#         for _k in dir(_tmp_cfgModule):
#             if not _k.startswith("__"):
#                 setattr(_tmp_thisModule, _k, getattr(_tmp_cfgModule, _k))
#     else:
#         print("Cannot load configuration. Configuration variable ", _tmp_cfgFile, " does not exist.", file=sys.stderr)
# else:
#     print("Cannot load configuration. Environment variable ", _tmp_ENVVARNAME_CFG, " not defined.", file=sys.stderr)

# Remove all temporary variables from this moduel.
for _k in dir(sys.modules[__name__]):
    if _k.startswith("_tmp_"):
        delattr(sys.modules[__name__], _k)

def log(logFunc):
    thisModule = sys.modules[__name__]    
    logFunc("")
    
    # If the environment variable is set to print the environment.
    pe=os.environ.get("PRINTENV")
    if pe and (pe.lower() in ("1", "true", "yes")):
        logFunc("*** ENV ******************************************")
        for k in os.environ.keys():
            logFunc(f"ENV.{k}: {os.environ.get(k)}")
        logFunc("*** ENV ******************************************")
        logFunc("")    
    
    # Print the configuration    
    logFunc("*** CFG ******************************************")
    for e in dir(thisModule):
        if e[0] == "_":
            continue
        a = getattr(thisModule, e, None)
        if not a:
            continue        
        if type(a) == str:
            logFunc(f"CFG.{e}: {a}")
        elif type(a) == dict:
            logFunc(f"CFG.{e}:")
            for k in a.keys():
                logFunc(f"  {k}: {a[k]}")
        elif type(a) == list:
            logFunc(f"CFG.{e}:")
            cnt=i
            for v in a:
                logFunc(f" {i}: {v}")
                i+=1
    logFunc("*** CFG ******************************************")
    logFunc("")    


import json
import logging

import dtiot_d2c.cli.CFG as CFG
import dtiot_d2c.cli.profile_webapp as profile_webapp
import dtiot_d2c.cli.utils as utils
from dtiot_d2c.cli.cli_command import (CmdLineInterface,
                                       ConfigCmdLineInterface,
                                       ProfileCmdLineInterface)
from dtiot_d2c.cli.config import Configuration
from dtiot_d2c.cli.profile import Profile, Profiles

log = logging.getLogger(__name__)
#----------------------------------------------------------------
def _editProfile(config, profiles, profile:Profile, isNew=True):

    if isNew:
        prompt = "add profile - "
    else:
        prompt = "modify profile - "

    for e in CFG.PROFILE_CONFIG["elements"]:
        #log.debug(f"e: {e}")
        if (n := e.get("name", None)):
            dv = profile.getElement(n)
        else:
            dv = str(e["default"])
            
        v = utils.getUserInput(prompt=prompt, title=e["label"], defaultValue=dv, requiresInput=True)
        profile.setElement(e["name"], v, e["type"])

    # Build the filename for the profile and verify if it exists
    if isNew:
        fname = profile.buildFilename(CFG.PROFILE_CONFIG)        

        if profiles.getProfile(fname):
            print("error: Profile file %s NOT unique. Choose a different combination of account and project." % (file))
            return          

        profile.setFile(fname)

    profile.write(directory=config.getProfilesDirectory(), file=profile.getFile())

#################################################################
# ListProfiles
#################################################################

class ListProfilesCLI(ConfigCmdLineInterface):
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Lists the profiles in the configuration."
    
    def addCmdLineArgs(self, argsParser):

        argsParser.add_argument("-s", "--select", metavar='<select>', 
                                help="Comma separated list of elements to be selected.")

        argsParser.add_argument("-f", "--filter", metavar='<filter>', dest="filter",
                                help="Filters the  profiles by its filename with the pattern \"*<filter>*\".")

        argsParser.add_argument("--separator-char", metavar='<char>', dest="sepchar", default=";",
                                help="Character to separate the fields in the select response. Default value is ;")        

    def main(self, cmdargs=None):
        log.debug("=> ListProfilesCLI()")

        profiles = Profiles(config=self.getConfig())

        d = {}

        #  If a filter is defined.
        if cmdargs.filter and len(cmdargs.filter) > 0:
            for key in profiles.getProfileKeys():
                profile = profiles.getProfile(key)
                if profile.getFile().lower().find(cmdargs.filter) > -1:
                    d[key] =  profile.getAsDict()
                    d[key]["file"] = key # The file is used as a key in the profiles dictionary
        else:
            for key in profiles.getProfileKeys(): 
                d[key] = profiles.getProfile(key).getAsDict() 
                d[key]["file"] = key # The file is used as a key in the profiles dictionary

        if not cmdargs.select:            
            print(json.dumps(d, indent=4))
        else:
            select_fields = utils.splitString(cmdargs.select, ",")
            for (k, prf) in d.items():
                a = []
                for sf in select_fields:
                    v = prf.get(sf, None)
                    a.append(str(v))
                line = cmdargs.sepchar.join(a)
                print("%s" % (line))

#################################################################
# PrintProfile
#################################################################

class PrintProfileCLI(ConfigCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Prints a profile of the configuration."
    
    def addCmdLineArgs(self, argsParser):

        argsParser.add_argument("-f", "--file", metavar='<file>', required=True,
                                help="Filename of the profile to be printed.")

        argsParser.add_argument("-s", "--select", metavar='<select>', 
                                help="Comma separated list of elements to be selected.")

        argsParser.add_argument("--separator-char", metavar='<char>', dest="sepchar", default=";",
                                help="Character to separate the fields in the select response. Default value is ;")        

    #def main(self, cmdargs=None, file=None, select=None, sepChar=None, config=None):
    def main(self, cmdargs=None, profile=None, select=None, sepChar=None):
        log.debug("=> PrintProfileCLI()")

        if not profile:
            profiles = Profiles(config=self.getConfig())
            file = cmdargs.file
            profile = profiles.getProfile(file)
    
            if not profile:
                raise Exception("Profile %s not available." % (file))

        if not select and cmdargs:
            select = cmdargs.select

        if not sepChar and cmdargs:
            sepChar = cmdargs.sepchar
        if not sepChar:
            sepChar = ";"

        d = profile.getAsDict()
        d["file"] = profile.getFile()

        if not select:        
            print(json.dumps(d, sort_keys=False, indent=4))

        else:
            a = []
            for s in utils.splitString(select, ","):
                if s in d.keys():
                    a.append(str(d[s]))
            line = sepChar.join(a)
            print("%s" % (line))

#################################################################
# NewProfile
#################################################################

class NewProfileCLI(ConfigCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Creates a new profile by collecting required infos from user input."
    
    def addCmdLineArgs(self, argsParser):

        argsParser.add_argument("--copy-from-profile", metavar='<profile name>', 
                                dest="copy_from_profile",
                                help="filiename of the profile which values shall be used to initialize the new profile.")

    def main(self, cmdargs=None, config=None):
        log.debug("=> NewProfileCLI()")
    
        if not config:
            config = self.getConfig()


        profiles = Profiles(config=config)

        # Create an empty profile
        profile = Profile(profileCfg=CFG.PROFILE_CONFIG)
        
        if cmdargs and cmdargs.copy_from_profile:
            #log.debug(f"cmdargs.copy_from_profile: {cmdargs.copy_from_profile}")
            cpFromPrf = profiles.getProfile(cmdargs.copy_from_profile)
            if not cpFromPrf:
               raise Exception("Copy from profile %s does not exist!" % (cmdargs.copy_from_profile))
            profile.copyFrom(cpFromPrf)
            #log.debug(f"profile: {profile}")

        _editProfile(config, profiles, profile, isNew=True)

        return profile

#################################################################
# EditProfile
#################################################################
class EditProfileCLI(ConfigCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Modifies profile by collecting user input from console."
    
    def addCmdLineArgs(self, argsParser):
        argsParser.add_argument("-f", "--file", metavar='<file>', required=True,
                                help="Filename of the profile which shall be modified.")

    #def main(self, cmdargs=None, configFilePath=None, file=None):
    def main(self, cmdargs=None):
        log.debug("=> EditProfileCLI()")

        config = self.getConfig()
        profiles = Profiles(config=config)
        profile = profiles.getProfile(cmdargs.file)

        if not profile:
            raise Exception("Profile %s does not exist!" % (cmdargs.file))

        _editProfile(config, profiles, profile, isNew=False)

#################################################################
# Activate Profile
#################################################################
class ActivateProfileCLI(ConfigCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Acticates a profile to be used as default."
    
    def addCmdLineArgs(self, argsParser):

        argsParser.add_argument("-f", "--file", metavar='<file>', required=True,
                                help="Filename of the profile to activate.")

    def main(self, cmdargs=None, config=None, profile=None):
        log.debug("=> ActivateProfileCLI()")

        if not config:
            config = self.getConfig()

        if not profile:            
            profiles = Profiles(config=config)
            profile = profiles.getProfile(cmdargs.file)

            if not profile:
                raise Exception("Profile %s does not exist!" % (cmdargs.file))

        config.setActiveProfileFile(profile.getFile())
        config.write()

#################################################################
# Print Activated Profile
#################################################################
class PrintActivatedProfileCLI(ProfileCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Prints the current activated profile."
    
    def addCmdLineArgs(self, argsParser):
        argsParser.add_argument("-s", "--select", metavar='<select>', 
                                help="Comma separated list of elements to be selected.")

        argsParser.add_argument("--separator-char", metavar='<char>', dest="sepchar", default=";",
                                help="Character to separate the fields in the select response. Default value is ;")        

    def main(self, cmdargs=None):
        log.debug("=> PrintActivatedProfileCLI()")

        profile = self.getProfile()

        if not profile:
            raise Exception("No profile activated")

        # Get the activate profile name and print it.
        PrintProfileCLI().main(profile=profile,
                               select=cmdargs.select, 
                               sepChar=cmdargs.sepchar)

#################################################################
# Start Profile Manager Gui
#################################################################
class StartProfileManagerGuiCLI(ConfigCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Starts the profile manager gui."
    
    def addCmdLineArgs(self, argsParser):
        pass
                
    def main(self, cmdargs=None):
        from dtiot_d2c.cli.prfmangui.ProfileManagerGui import ProfileManagerGui
        prfManGui = ProfileManagerGui(self.getConfig())
        prfManGui.run()

#################################################################
# Start Profile Manager application in a WebBrowser
#################################################################
class StartProfileManagerWebBrowserCLI(ConfigCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Starts the profile manager application in a web browser."
    
    def addCmdLineArgs(self, argsParser):
        argsParser.add_argument("-u", "--url", metavar='<url>', default="http://127.0.01:3000/login",
                                help="URL on which the profile manager web application can be accessed.")

    def main(self, cmdargs=None):

        profile_webapp.open_in_browser(webapp_url=cmdargs.url)

        #from dtiot_d2c.cli.prfmangui.ProfileManagerGui import ProfileManagerGui
        #prfManGui = ProfileManagerGui(self.getConfig())
        #prfManGui.run()



Device01Cfg = {
    "name":"device01-%RUNID%",
    "create_props" : {
        "manufacturer":"PSsystec",
        "model":"Smartbox Mini",
        "sub_model":"NB-IoT",
        "iccid": "89374121317959919924",
        "hardware_version":"2024.1 PC r1",
        "country":"DE",
        "description":"Das ist eine Beschreibung.",
        "label":"this is a label" ,
        "device_type":"Pretty Device" ,
        "device_name":"This is the name of the device",
        "firmware_version":"fw 1.0.1",
        "software_version":"sw 2.0.1",
        "os_version":"os 3.0.1",
        "location":"geo:25.245470,51.454009" ,
        "protocols":["LWM2M"] ,
        "uplink_properties": {"address":"Kuckhoffstr 114A, 13156 Berlin"},
        "device_properties": {"lastMaintenance":"25.3.2025 1300", "prop2":"value2"},
        "credentials_id": "device01-%RUNID%",
        "credentials_secret": "Q3hhc21FTUw3NFNiUlIzUw=="
    },
    "modify_props" : {
        "manufacturer":"x PSsystec",
        "model":"x Smartbox Mini",
        "sub_model":"x NB-IoT",
        "iccid": "99999999317959919924",
        "hardware_version":"x 2024.1 PC r1",
        "country":"XX",
        "description":"x Das ist eine Beschreibung.",
        "label":"x this is a label" ,
        "device_type":"x Pretty Device" ,
        "device_name":"Hallo die Waldfee device_name",
        "firmware_version":"x fw 1.0.1",
        "software_version":"x sw 2.0.1",
        "os_version":"x os 3.0.1",
        "location":"geo:88.999999,88.999999" ,
        "protocols":["LWM2M"] ,
        "uplink_properties": {"address":"xx Kuckhoffstr 114A, 13156 Berlin", "location":"Berlin"},
        "device_properties": {"lastMaintenance":"25.3.2025 1300","Status": "Out of order"},

        #Update lwm2m cridentials is not supported yet via d2ccli
        #"credentials_id": "x device01-%RUNID%",
        #"credentials_secret": "x Q3hhc21FTUw3NFNiUlIzUw=="
    }    
} 
Device02Cfg = {
    "name":"device02-%RUNID%",
    "create_props" : {
        "manufacturer":"PSsystec",
        "model":"Smartbox Mini",
        "sub_model":"NB-IoT",
        "iccid": "89374121317959919924",
        "hardware_version":"2024.1 PC r1",
        "country":"DE",
        "description":"Das ist eine Beschreibung.",
        "label":"this is a label" ,
        "device_type":"Pretty Device" ,
        "firmware_version":"fw 1.0.1",
        "software_version":"sw 2.0.1",
        "os_version":"os 3.0.1",
        "location":"geo:25.245470,51.454009" ,
        "protocols":["LWM2M"] ,
        "uplink_properties": {"address":"Kuckhoffstr 114A, 13156 Berlin"},
        "device_properties": {"lastMaintenance":"25.3.2025 1300"},
        "credentials_id": "device01-%RUNID%",
        "credentials_secret": "Q3hhc21FTUw3NFNiUlIzUw=="
    }
} 

Application01Cfg = {
    "name":"application01-%RUNID%",
    "create_props" : {
        "application_type":"webHook",
        "urls": ["https://api.scs.iot.telekom.com/message-monitor"],
        "connection_properties": {"Auth":"dfadfdsafasdf", "another-token":"88888888"}
    }, 
    "modify_props" : {    
        "urls": ["https://dev.scs.iot.telekom.com/scs-callback-dummy"],
        "connection_properties": {"Auth":"88888888", "another-token":"dfadfdsafasdf"}
    }
}

DeviceGroup01Cfg = {
    "name":"devicegroup01-%RUNID%", 
    "create_props" : {
        "labels":{"deviceType":"SDI People Counter"},
        "device_ids": [],
        "application_ids": []
    },
    "modify_props":{
        "labels":{"deviceType":"SDI People Counter, ELSYS", "street":"Kuckhoffstr."},
        "device_ids": [],
        "application_ids": []
    }
}

DeviceGroup02Cfg = {
    "name":"devicegroup02-%RUNID%", 
    "create_props" : {
        "labels":{"deviceType":"SDI People Counter"},
        "device_ids": [],
        "application_ids": []
    },
    "application_1": {
        "name":"devicegroup02-app1-%RUNID%",
        "create_props":{
            "urls": ["https://api.scs.iot.telekom.com/message-monitor"],
            "labels": {"customerKey":"CUST-001"},        
        }
    },
    "application_2":{
        "name":"devicegroup02-app2-%RUNID%",
        "create_props":{
            "urls": ["https://dev.scs.iot.telekom.com/scs-callback-dummy"],
            "labels": {"customerKey":"CUST-099", "costCenter":"999888"},        
        }
    },
    "device_1": Device01Cfg,
    "device_2": Device02Cfg
}


#!/usr/bin/env python3

import json
from fastapi import Request

import logging
log = logging.getLogger(__name__)

def copyHeaders(headers, namesToLower=False):
    ret = {}
    for k in headers.keys():
        if namesToLower:
            ret[k.lower()] = headers[k]
        else:
            ret[k] = headers[k]
    return ret

def logRequest(request:Request, body=None):
    log.debug(">>> HEADER")
    for h in request.headers.keys():
        log.debug("%s : %s" % (h, request.headers[h]))
    log.debug("")
    for qp in request.query_params.keys():
        log.debug("%s : %s" % (qp, request.query_params[qp]))

    if body:
        log.debug("")
        log.debug(">>> BODY")
        if type(body) == dict:
            log.debug(json.dumps(body, indent=4))
        elif type(body) == str:
            log.debug(body)
        else:
            log.debug(str(body))
    log.debug(">>>")




import json
import logging
import os
import sys
import time
from typing import Callable, ClassVar, List

from pydantic import BaseModel, Field

import dtiot_d2c.d2c.test.utils as tutils
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.application import Application
from dtiot_d2c.d2c.device.consts import *
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.dmo import ApiClient
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
            
def print_flushed(s:str, end=None):
    print(s, end=end)
    sys.stdout.flush() 

class TestCase:
    def __init__(self, api: ApiClient, cfg:dict):
        self.api = api
        self.cfg = cfg

    def get_name(self):
        return "TestCase"
    
    def run(self):
        raise Exception("run() function requires implementation.")
    
    def verify(self):
        return None
    
class TestSet:
    def __init__(self, api: ApiClient, name:str):
        self.api = api
        self.name = name
    
    def get_sample_configuration_varname(self)->str:
        raise Exception("get_sample_configuration_varname() function requires implementation by child class.")

    def get_test_cases(self, cfg:dict)->List[TestCase]:
        return []    
    
    def test_pre_cleanup(self, cfg:dict)->str:
        return None

    def test_post_cleanup(self, cfg:dict)->str:
        return None
        
    def run_all_tests(self, cfg:dict, **kwargs):
        number_of_test_loops:int = kwargs.get("number_of_test_loops", 1)
        skip_pre_cleanup:bool = kwargs.get("skip_pre_cleanup", False)
        skip_post_cleanup:bool = kwargs.get("skip_post_cleanup", False)
                
        if not skip_pre_cleanup:
            tutils.print_BOLD(f"*** PRE CLEANING UP ...")
            errmsg = None
            try:
                errmsg=self.test_pre_cleanup(cfg)
                #if not errmsg:
                #    tutils.print_OK()                

            except Exception as ex:
                errmsg = str(ex)

            if errmsg:
                tutils.print_ERROR(f"An error occured during pre cleaning up.")
                tutils.print_ERROR(errmsg)                
                return
                
            print("")
        else:
            tutils.print_BOLD(f"*** SKIPPING PRE CLEANING UP ...")
            
        ok_count = 0
        finished_with_errmsg = None
        start_time = time.time()

        for i in range(number_of_test_loops):

            if number_of_test_loops > 1:
                tutils.print_BOLD(f"*** TESTRUN {i + 1} of {number_of_test_loops} **********************************", flush=True)

            errmsg = None
            try:
                for tc in self.get_test_cases(cfg):
                    tc_name = tc.get_name()
                    tutils.print_BOLD(f"RUNNING {tc_name} ...")
                    tc.run()
                    print("")
                    tutils.print_BOLD(f"VERIFYING {tc_name} ...")                    
                    tc.verify()
                    print("")

            except Exception as ex:
                errmsg = str(ex)

            if errmsg:
                tutils.print_ERROR(f"An error occured. Stopping test during the {i+1}. run.")
                tutils.print_ERROR(errmsg)
                finished_with_errmsg = errmsg
                break
            else:
                ok_count += 1
                
        end_time = time.time()

        duration_total = end_time - start_time
        duration_per_run = (duration_total / ok_count) if ok_count > 0 else duration_total

        if not skip_post_cleanup:
            tutils.print_BOLD(f"*** POST CLEANING UP ...")
            errmsg = None
            try:
                errmsg=self.test_post_cleanup(cfg)
                #if not errmsg:
                #    tutils.print_OK()
                    
            except Exception as ex:
                errmsg = str(ex)

            if errmsg:
                tutils.print_ERROR(f"An error occured during post cleaning up.")
                tutils.print_ERROR(errmsg)             

            print("")
        else:
            tutils.print_BOLD(f"*** SKIPPING POST CLEANING UP ...")
            
        if not finished_with_errmsg:
            print_func = tutils.print_OK
            print_func(f"*** END OK ***")
        else:
            print_func = tutils.print_ERROR
            print_func(f"*** END ERROR ***")
            print_func(f"Error message:       {finished_with_errmsg}")
    
        print_func(f"Number of test runs: {ok_count}")
        print_func(f"Total time:          {duration_total:.4f} seconds ")
        print_func(f"Time per test run:   {duration_per_run:.4f} seconds")
class TestCLI_Base(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()
    
    def addCmdLineArgs(self, argsParser):
        argsParser.add_argument("-i", "--test-run-id", metavar="<string>", 
                                dest="run_id", required=True,
                                help="Id of the test run.")

        argsParser.add_argument("-cf", "--test-configuration-file", metavar="<file>", 
                                dest="config_file", required=False,
                                help="Python file which contains the configuration of the test.")

        argsParser.add_argument("-cd", "--test-configuration-dict", metavar="<variable-name>", 
                                dest="config_varname", required=False,
                                help="Name of the dictionary variable within the configuration file with the configuration.")

        argsParser.add_argument("-spr", "--skip-pre-cleanup", dest="skip_pre_cleanup", action="store_const", 
                                const=True, default=False,
                                help="If set, the pre cleanup function is skipped.")

        argsParser.add_argument("-spo", "--skip-post-cleanup", dest="skip_post_cleanup", action="store_const", 
                                const=True, default=False,
                                help="If set, the post cleanup function is skipped.")

        argsParser.add_argument("-n", "--number-of-test-runs", metavar="<integer>", 
                                dest="n", type=int, default=1,
                                help="Number of test runs. Default is 1.")

    def get_test_set(self, api: ApiClient)->TestSet:
        raise Exception("get_test_set() requires implementation in child class.")
        
    def main(self, cmdargs=None):

        test_set:TestSet = self.get_test_set(self.apiConn)
        run_id = cmdargs.run_id

        # If no configuration file is defined use the sample configuration as default.
        config_file = cmdargs.config_file
        config_varname = cmdargs.config_varname

        if not config_file:
            config_file = f"{os.path.dirname(__file__)}/sample_test_configuration.py"
            config_varname = test_set.get_sample_configuration_varname()

        # Configuration file has been defined but no name of the configuration variable
        # within the configuration file is defined. 
        elif not config_varname:
            raise Exception(f"Command line option --test-configuration-dict required.")
        
        if cmdargs.n <= 0:
            raise Exception(f"Parameters -n or --number-of-test-runs shall be greater than 0.")
        else:
            n = cmdargs.n

        # Load the test configuration            
        cfg = tutils.load_test_configuration(run_id, config_file, config_varname)

        # Run all tests
        test_set.run_all_tests(cfg,
                               number_of_test_loops=cmdargs.n,
                               skip_pre_cleanup=cmdargs.skip_pre_cleanup,
                               skip_post_cleanup=cmdargs.skip_post_cleanup
                              )



import json
import logging
from argparse import ArgumentParser

import dtiot_d2c.d2c as d2c
import dtiot_d2c.dmo as dmo
from dtiot_d2c.cli.cli_command import ApiConnCmdLineInterface
from dtiot_d2c.d2c import CMDARGS
from dtiot_d2c.d2c import utils as utils
from dtiot_d2c.d2c.device import Device
from dtiot_d2c.d2c.device.consts import *
from dtiot_d2c.d2c.get_response import GetResponse
from dtiot_d2c.dmo import dmoCLI as dmocli

log = logging.getLogger(__name__)
            
#####################################################################################
class AddDeviceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds a device."

    def addCmdLineArgs(self, argsParser):

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n", required=True)

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de")  # descripton
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-la")  # label
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dty") # device type
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dna") # device name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-fv")  # firmware version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-sv")  # software version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ov")  # os version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-lo")  # location

        if not self.hasPreDefinedCmdArg("--protocol"):
            self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-pr")  # protocols

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ma")  # manufacturer
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-mo")  # model
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-smo") # sub model
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ic")  # iccid
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-hv")  # hardware version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cy")  # country code
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-up")  # uplink properties
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dp")  # device properties

        # LWM2M
        #if not self.hasHideCmdArg("--lwm2m"):

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cid")  # credentials id
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-csec") # credentials secret
        
    def main(self, cmdargs=None):
        
        if cmdargs.protocol in ["LoRaWAN", "UDP", "MQTT", "DTLS CoAP", "CoAP", "HTTP"]:
            raise Exception(f"Protocol {cmdargs.protocol} not supported yet.")

        if not cmdargs.credentials_id:
            cmdargs.credentials_id = cmdargs.name

        uplink_properties = json.loads(cmdargs.uplink_properties) if cmdargs.uplink_properties else None
        device_properties = json.loads(cmdargs.device_properties) if cmdargs.device_properties else None
        
        response = Device.create(self.apiConn, 
                                 origin=cmdargs.origin or self.apiConn.origin,
                                 name=cmdargs.name,
                                 iccid=getattr(cmdargs, "iccid", ""),
                                 description=cmdargs.description, 
                                 label=getattr(cmdargs, "label", None),
                                 device_type=getattr(cmdargs, "device_type", None),
                                 device_name=getattr(cmdargs, "device_name", None),
                                 firmware_version=getattr(cmdargs, "firmware_version", None),
                                 software_version=getattr(cmdargs, "software_version", None),
                                 os_version=getattr(cmdargs, "os_version", None),
                                 location=getattr(cmdargs, "location", None),
                                 protocols=[cmdargs.protocol],                                
                                 manufacturer=getattr(cmdargs, "manufacturer", None),
                                 model=getattr(cmdargs, "model", None),
                                 sub_model=getattr(cmdargs, "sub_model", None),
                                 country=getattr(cmdargs, "country", None),
                                 hardware_version=getattr(cmdargs, "hardware_version", None),
                                 uplink_properties=uplink_properties,
                                 device_properties=device_properties,
                                 credentials_id=getattr(cmdargs, "credentials_id", None),                                
                                 credentials_secret=getattr(cmdargs, "credentials_secret", None),
                                )
        
        GetResponse(response=response).print()

class AddLwm2mDeviceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Adds a lwm2m device."

    def addCmdLineArgs(self, argsParser):

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n", required=True)
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cid", required=True)  # credentials id
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-csec", required=True) # credentials secret

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de")  # descripton
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-la")  # label
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dty") # device type
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dna") # device name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-fv")  # firmware version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-sv")  # software version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ov")  # os version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-lo")  # location

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ma")  # manufacturer
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-mo")  # model
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-smo") # sub model
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ic")  # iccid
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-hv")  # hardware version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cy")  # country code
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-up")  # uplink properties
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dp")  # device properties


    def main(self, cmdargs=None):
        uplink_properties = json.loads(cmdargs.uplink_properties) if cmdargs.uplink_properties else None
        device_properties = json.loads(cmdargs.device_properties) if cmdargs.device_properties else None
        
        response = Device.create(self.apiConn, 
                                 origin=cmdargs.origin or self.apiConn.origin,
                                 name=cmdargs.name,
                                 iccid=getattr(cmdargs, "iccid", ""),
                                 description=cmdargs.description, 
                                 label=getattr(cmdargs, "label", None),
                                 device_type=getattr(cmdargs, "device_type", None),
                                 device_name=getattr(cmdargs, "device_name", None),
                                 firmware_version=getattr(cmdargs, "firmware_version", None),
                                 software_version=getattr(cmdargs, "software_version", None),
                                 os_version=getattr(cmdargs, "os_version", None),
                                 location=getattr(cmdargs, "location", None),
                                 protocols=["LWM2M"],                                
                                 manufacturer=getattr(cmdargs, "manufacturer", None),
                                 model=getattr(cmdargs, "model", None),
                                 sub_model=getattr(cmdargs, "sub_model", None),
                                 country=getattr(cmdargs, "country", None),
                                 hardware_version=getattr(cmdargs, "hardware_version", None),
                                 uplink_properties=uplink_properties,
                                 device_properties=device_properties,
                                 credentials_id=getattr(cmdargs, "credentials_id", None),                                
                                 credentials_secret=getattr(cmdargs, "credentials_secret", None),
                                )
        
        GetResponse(response=response).print()
        
#####################################################################################
class UpdateDeviceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Updates the attributes of a device."

    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-de")  # descripton
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-la")  # label
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dty") # device type
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dna") # device name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-fv")  # firmware version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-sv")  # software version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ov")  # os version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-hv")  # hardware version
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-lo")  # location
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-pr")  # protocols
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ma")  # manufacturer
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-mo")  # model
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-smo") # sub model
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ic")  # iccid
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cy")  # country
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-up")  # uplink properties
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-dp")  # device properties

        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-cid")  # credentials id
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-csec") # credentials secret

    def main(self, cmdargs=None):
        
        if cmdargs.protocol in ["LoRaWAN", "UDP", "MQTT", "DTLS CoAP", "CoAP", "HTTP"]:
            raise Exception(f"Protocol {cmdargs.protocol} not supported yet.")

        if not cmdargs.name and not cmdargs.id:
            raise Exception(f"--name or --id option required to identify device.")            

        uplink_properties = json.loads(cmdargs.uplink_properties) if cmdargs.uplink_properties else None
        device_properties = json.loads(cmdargs.device_properties) if cmdargs.device_properties else None
        
        response = Device.update(self.apiConn, 
                                 origin=cmdargs.origin or self.apiConn.origin,
                                 name=cmdargs.name,
                                 id=cmdargs.id,                                 
                                 iccid=getattr(cmdargs, "iccid", ""),
                                 description=cmdargs.description, 
                                 label=getattr(cmdargs, "label", None),
                                 device_type=getattr(cmdargs, "device_type", None),
                                 device_name=getattr(cmdargs, "device_name", None),
                                 firmware_version=getattr(cmdargs, "firmware_version", None),
                                 software_version=getattr(cmdargs, "software_version", None),
                                 os_version=getattr(cmdargs, "os_version", None),
                                 location=getattr(cmdargs, "location", None),
                                 protocols=[cmdargs.protocol],                                
                                 manufacturer=getattr(cmdargs, "manufacturer", None),
                                 model=getattr(cmdargs, "model", None),
                                 sub_model=getattr(cmdargs, "sub_model", None),
                                 country=getattr(cmdargs, "country", None),
                                 hardware_version=getattr(cmdargs, "hardware_version", None),
                                 uplink_properties=uplink_properties,
                                 device_properties=device_properties,
                                 credentials_id=getattr(cmdargs, "credentials_id", None),                                
                                 credentials_secret=getattr(cmdargs, "credentials_secret", None),
                                )
        if not response:
            raise Exception(f"Device {cmdargs.name or cmdargs.id} does not exist")
                
        GetResponse(response=response).print()

#####################################################################################
class GetDevicesCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the details of a single device or a list of devices."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-li")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-os")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-rci")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-s")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-f")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-sc")
                
    def main(self, cmdargs=None):
        response = Device.get(self.apiConn,
                              origin=cmdargs.origin,
                              name=cmdargs.name,
                              id=cmdargs.id,
                              limit=cmdargs.limit,
                              offset=cmdargs.offset,
                              resultContentIndocator=cmdargs.rci, 
                              select=cmdargs.select,
                              format=cmdargs.format)

        if not response and (cmdargs.name or cmdargs.id):
            raise Exception(f"Device {cmdargs.name or cmdargs.id} does not exist")   

        GetResponse(response=response).print(cmdargs.sepchar)

#####################################################################################            
class DeleteDeviceCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Deletes a device object."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
                
    def main(self, cmdargs=None):
        obj = Device.delete(self.apiConn, 
                            origin=cmdargs.origin, 
                            id=cmdargs.id,
                            name=cmdargs.name)

        if obj:
            print(str(obj))            
            
#####################################################################################            
class InjectUplinkMessageCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Injects an uplink message for a device to d2c."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
        
        argsParser.add_argument("-con", "--content", metavar="<string>", dest="content",
                                required=False, 
                                help="Content to inject as message.")

        argsParser.add_argument("-conf", "--content-file", metavar="<file>", dest="content_file",
                                required=False, 
                                help="File which contains the content to inject as message.")

        argsParser.add_argument("-cont", "--content-type", metavar="<mime-type>", dest="content_type",
                                choices=["text/plain", "application/json"], default="text/plain",
                                required=False, 
                                help="Type of the content. Default is text/plain.")

    def main(self, cmdargs=None):
        content = None                
        
        if cmdargs.content:
            content = cmdargs.content 
        elif cmdargs.content_file:
            f = open(cmdargs.content_file, "r")
            content = f.read()
            f.close()
        else:
            raise Exception("Parameter --content or --content-file required.")                            
        
        #content = json.loads(content)
        response = Device.inject_uplink_message(self.apiConn,
                                                origin=cmdargs.origin,
                                                id=cmdargs.id,
                                                name=cmdargs.name,
                                                content=content,
                                                content_type=cmdargs.content_type
                                               )
        if response:
            GetResponse(response=response).print()

#####################################################################################            
class InjectDownlinkMessageCLI(ApiConnCmdLineInterface):

    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Injects a downlink message for a device to d2c."
    
    def addCmdLineArgs(self, argsParser):
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")
        
        argsParser.add_argument("-con", "--content", metavar="<string>", dest="content",
                                required=False, 
                                help="Content to inject as message.")

        argsParser.add_argument("-conf", "--content-file", metavar="<file>", dest="content_file",
                                required=False, 
                                help="File which contains the content to inject as message.")

        argsParser.add_argument("-cont", "--content-type", metavar="<mime-type>", dest="content_type",
                                choices=["text/plain", "application/json"], default="text/plain",
                                required=False, 
                                help="Type of the content. Default is text/plain.")

    def main(self, cmdargs=None):
        content = None                
        
        if cmdargs.content:
            content = cmdargs.content 
        elif cmdargs.content_file:
            f = open(cmdargs.content_file, "r")
            content = f.read()
            f.close()
        else:
            raise Exception("Parameter --content or --content-file required.")                            
        
        #content = json.loads(content)
        response = Device.inject_downlink_message(self.apiConn,
                                                  origin=cmdargs.origin,
                                                  id=cmdargs.id,
                                                  name=cmdargs.name,
                                                  content=content,
                                                  content_type=cmdargs.content_type
                                                 )
        if response:
            GetResponse(response=response).print()
            
#####################################################################################
class GetDeviceMessagesCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Returns the messages related to a device."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n") # Device name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i") # Device id
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ms", required=True) # Message store
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-mn")  # Message name

        argsParser.add_argument("-la", "--last", dest="last", action="store_const", const=True, default=False,
                                help="If set only the last message is returned from the store.")
                
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-li")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-os")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-s")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-f")
        self.addCmdLineArgsFromTemplate(dmocli.cmdArgTemplates, argsParser, "-sc")
                
    def main(self, cmdargs=None):
        
        msg_store:MessageStore = None
        if cmdargs.msg_store in ["uplink-inbound", "uli"]:
            msg_store = INBOUND_UPLINK_MESSAGE_STORE
        elif cmdargs.msg_store in ["uplink-outbound", "ulo"]:
            msg_store = OUTBOUND_UPLINK_MESSAGE_STORE
            raise Exception("Outbound uplink messages are not supported yet.")
        if cmdargs.msg_store in ["downlink-inbound", "dli"]:
            msg_store = INBOUND_DOWNLINK_MESSAGE_STORE
        elif cmdargs.msg_store in ["downlink-outbound", "dlo"]:
            msg_store = OUTBOUND_DOWNLINK_MESSAGE_STORE
        
        response = Device.get_messages(self.apiConn,
                                       origin=cmdargs.origin,
                                       name=cmdargs.name,
                                       id=cmdargs.id,
                                       msg_store=msg_store,
                                       msg_name=cmdargs.message_name,
                                       last=cmdargs.last,
                                       limit=cmdargs.limit,
                                       offset=cmdargs.offset,
                                       select=cmdargs.select,
                                       format=cmdargs.format)
        
        GetResponse(response=response).print(cmdargs.sepchar)
        
#####################################################################################
class DeleteDeviceMessageCLI(ApiConnCmdLineInterface):
    
    def __init__(self):
        super().__init__()

    def get1LineHelp(self):
        return "Deletes a device message from a mesage store."
    
    def addCmdLineArgs(self, argsParser):
        
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-o")
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-n")  # Device name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-i")  # Device id
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-ms") # Message store
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-mn") # Message name
        self.addCmdLineArgsFromTemplate(CMDARGS, argsParser, "-mi") # Message id
                
    def main(self, cmdargs=None):
        
        msg_store:MessageStore = None
        if cmdargs.msg_store in ["uplink-inbound", "uli"]:
            msg_store = INBOUND_UPLINK_MESSAGE_STORE
        elif cmdargs.msg_store in ["uplink-outbound", "ulo"]:
            msg_store = OUTBOUND_UPLINK_MESSAGE_STORE
            raise Exception("Outbound uplink messages are not supported yet.")
        if cmdargs.msg_store in ["downlink-inbound", "dli"]:
            msg_store = INBOUND_DOWNLINK_MESSAGE_STORE
        elif cmdargs.msg_store in ["downlink-outbound", "dlo"]:
            msg_store = OUTBOUND_DOWNLINK_MESSAGE_STORE
        
        Device.delete_message(self.apiConn,
                              origin=cmdargs.origin,
                              name=cmdargs.name,
                              id=cmdargs.id,
                              msg_store=msg_store,
                              msg_name=cmdargs.message_name,
                              msg_id=cmdargs.message_id)
